#
# Copyright (c) 2018, 2022 Oracle and/or its affiliates. All rights reserved.
#
# Licensed under the Universal Permissive License v 1.0 as shown at
#  https://oss.oracle.com/licenses/upl/
#

from datetime import datetime
from decimal import Decimal
from functools import wraps
from logging import Logger
from struct import pack, unpack
from sys import version_info
from threading import Lock
from time import time
from warnings import simplefilter, warn

from dateutil import tz
from requests import adapters

from .exception import IllegalArgumentException


def enum(**enums):
    return type('Enum', (object,), enums)


def deprecated(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        simplefilter('default', DeprecationWarning)
        warn("Call to deprecated function {}.".format(func.__name__),
             DeprecationWarning)
        return func(*args, **kwargs)

    return wrapper


def synchronized(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        with self.lock:
            return func(self, *args, **kwargs)

    return wrapper


class ByteInputStream(object):
    """
    The ByteInputStream provides methods to get data with different type from
    a bytearray.
    """

    def __init__(self, content):
        self._content = content
        self._offset = 0

    def get_offset(self):
        return self._offset

    def read_boolean(self):
        res = bool(self.read_byte())
        return res

    def read_byte(self):
        res = self._content[self._offset]
        self._offset += 1
        if res > 127:
            return res - 256
        else:
            return res

    def read_float(self):
        buf = bytearray(8)
        self.read_fully(buf)
        res, = unpack('>d', buf)
        return res

    def read_fully(self, buf, start=0, end=None):
        if end is None:
            end = len(buf)
        for index in range(start, end):
            buf[index] = self._content[self._offset]
            self._offset += 1

    def read_int(self):
        buf = bytearray(4)
        self.read_fully(buf)
        res, = unpack('>i', buf)
        return res

    def read_long(self):
        buf = bytearray(8)
        self.read_fully(buf)
        res, = unpack('>q', buf)
        return res

    def read_short_int(self):
        buf = bytearray(2)
        self.read_fully(buf)
        res, = unpack('>h', buf)
        return res

    def set_offset(self, offset):
        self._offset = offset


class ByteOutputStream(object):
    """
    The ByteOutputStream provides methods to write data with different type into
    a bytearray.
    """

    def __init__(self, content):
        self._content = content

    def get_offset(self):
        return len(self._content)

    def write_boolean(self, value):
        val_s = pack('?', value)
        self.write_value(val_s)

    def write_byte(self, value):
        val_s = pack('B', value)
        self.write_value(val_s)

    def write_bytearray(self, value, start=0, end=None):
        if end is None:
            end = len(value)
        for index in range(start, end):
            self._content.append(value[index])

    def write_float(self, value):
        val_s = pack('>d', value)
        self.write_value(val_s)

    def write_int(self, value):
        val_s = pack('>i', value)
        self.write_value(val_s)

    def write_int_at_offset(self, offset, value):
        val_s = pack('>i', value)
        val_b = bytearray(val_s)
        for index in range(len(val_b)):
            self._content[offset + index] = val_b[index]

    def write_short_int(self, value):
        val_s = pack('>h', value)
        self.write_value(val_s)

    def write_value(self, value):
        val_b = bytearray(value)
        self.write_bytearray(val_b)


class CheckValue(object):

    @staticmethod
    def check_boolean(data, name):
        if data is not True and data is not False:
            raise IllegalArgumentException(name + ' must be True or False.')

    @staticmethod
    def check_dict(data, name):
        if data is not None and not isinstance(data, dict):
            raise IllegalArgumentException(name + ' must be a dict.')

    @staticmethod
    def check_float_gt_zero(data, name):
        if not CheckValue.is_digit(data) or data <= 0.0:
            raise IllegalArgumentException(
                name + ' must be a positive digital number. Got:' + str(data))

    @staticmethod
    def check_int(data, name):
        if not (CheckValue.is_int(data) or CheckValue.is_long(data)):
            raise IllegalArgumentException(
                name + ' must be an integer. Got:' + str(data))

    @staticmethod
    def check_int_ge_zero(data, name):
        if (not (CheckValue.is_int(data) or CheckValue.is_long(data)) or
                data < 0):
            raise IllegalArgumentException(
                name + ' must be an integer that is not negative. Got:' +
                str(data))

    @staticmethod
    def check_int_gt_zero(data, name):
        if (not (CheckValue.is_int(data) or CheckValue.is_long(data)) or
                data <= 0):
            raise IllegalArgumentException(
                name + ' must be an positive integer. Got:' + str(data))

    @staticmethod
    def check_list(data, name):
        if not isinstance(data, list):
            raise IllegalArgumentException(name + ' must be a list.')

    @staticmethod
    def check_logger(data, name):
        if data is not None and not isinstance(data, Logger):
            raise IllegalArgumentException(name + ' must be a Logger.')

    @staticmethod
    def check_not_none(data, name):
        if data is None:
            raise IllegalArgumentException(name + ' must be not-none.')

    @staticmethod
    def check_str(data, name, allow_none=False):
        if (not allow_none and data is None or data is not None and
                (not CheckValue.is_str(data) or len(data) == 0)):
            raise IllegalArgumentException(
                name + ' must be a string that is not empty.')

    @staticmethod
    def is_digit(data):
        if (CheckValue.is_int(data) or CheckValue.is_long(data) or
                isinstance(data, float) or isinstance(data, Decimal)):
            return True
        return False

    @staticmethod
    def is_int(data):
        if ((version_info.major == 2 and isinstance(data, int) or
             version_info.major == 3 and isinstance(data, int)) and
                -pow(2, 31) <= data < pow(2, 31)):
            return True
        return False

    @staticmethod
    def is_long(data):
        if ((version_info.major == 2 and isinstance(data, (int, long)) or
             version_info.major == 3 and isinstance(data, int)) and
                not CheckValue.is_int(data) and
                -pow(2, 63) <= data < pow(2, 63)):
            return True
        return False

    @staticmethod
    def is_overlong(data):
        if ((version_info.major == 2 and isinstance(data, long) or
             version_info.major == 3 and isinstance(data, int)) and
                (data < -pow(2, 63) or data >= pow(2, 63))):
            return True
        return False

    @staticmethod
    def is_str(data):
        if (version_info.major == 2 and isinstance(data, (str, unicode)) or
                version_info.major == 3 and isinstance(data, str)):
            return True
        return False


class Consistency(object):
    """
    Set the consistency for read requests.
    """
    ABSOLUTE = 0
    """
    Set Consistency.ABSOLUTE to use absolute consistency for read requests.
    """
    EVENTUAL = 1
    """
    Set Consistency.EVENTUAL to use eventual consistency for read requests.
    This is the default value for operations.
    """


class Empty(object):
    """
    Represents an EMPTY field value.
    """

    def __str__(self):
        return 'Empty'


class FieldRange(object):
    """
    FieldRange defines a range of values to be used in a
    :py:meth:`NoSQLHandle.multi_delete` operation, as specified in
    :py:meth:`MultiDeleteRequest.set_range`. FieldRange is only relevant
    if a primary key has multiple components because all values in the
    range must share the same shard key.

    FieldRange is used as the least significant component in a partially
    specified key value in order to create a value range for an operation that
    returns multiple rows or keys. The data types supported by FieldRange are
    limited to the atomic types which are valid for primary keys.

    The least significant component of a key is the first component of the key
    that is not fully specified. For example, if the primary key for a table is
    defined as the tuple (a, b, c), a FieldRange can be specified for "a" if the
    primary key supplied is empty. A FieldRange can be specified for "b" if the
    primary key supplied to the operation has a concrete value for "a" but not
    for "b" or "c".

    This object is used to scope a :py:meth:`NoSQLHandle.multi_delete`
    operation. The field_path specified must name a field in a table's primary
    key. The values used must be of the same type and that type must match the
    type of the field specified.

    Validation of this object is performed when is it used in an operation.
    Validation includes verifying that the field is in the required key and, in
    the case of a composite key, that the field is in the proper order relative
    to the key used in the operation.

    :param field_path: the path to the field used in the range.
    :type field_path: str
    :raises IllegalArgumentException: raises the exception if field_path is not
        a string.
    """

    def __init__(self, field_path):
        # Create a value based on a specific field.
        CheckValue.check_str(field_path, 'field_path')
        self._field_path = field_path
        self._start = None
        self._start_inclusive = False
        self._end = None
        self._end_inclusive = False

    def __str__(self):
        return ('{Path=' + self._field_path + ', Start=' + str(self._start) +
                ', End=' + str(self._end) + ', StartInclusive=' +
                str(self._start_inclusive) + ', EndInclusive=' +
                str(self._end_inclusive) + '}')

    def get_field_path(self):
        """
        Returns the name for the field used in the range.

        :returns: the name of the field.
        :rtype: str
        """
        return self._field_path

    def set_start(self, value, is_inclusive):
        """
        Sets the start value of the range to the specified value.

        :param value: the value to set.
        :type value: any
        :param is_inclusive: set to True if the range is inclusive of the value,
            False if it is exclusive.
        :type is_inclusive: bool
        :returns: self.
        :raises IllegalArgumentException: raises the exception if parameters are
            not expected type.
        """
        CheckValue.check_not_none(value, 'value')
        CheckValue.check_boolean(is_inclusive, 'is_inclusive')
        self._start = value
        self._start_inclusive = is_inclusive
        return self

    def get_start(self):
        """
        Returns the field value that defines lower bound of the range, or None
        if no lower bound is enforced.

        :returns: the start field value.
        """
        return self._start

    def get_start_inclusive(self):
        """
        Returns whether start is included in the range, i.e., start is less than
        or equal to the first field value in the range. This value is valid only
        if the start value is not None.

        :returns: True if the start value is inclusive.
        :rtype: bool
        """
        return self._start_inclusive

    def set_end(self, value, is_inclusive):
        """
        Sets the end value of the range to the specified value.

        :param value: the value to set.
        :type value: any
        :param is_inclusive: set to True if the range is inclusive of the value,
            False if it is exclusive.
        :type is_inclusive: bool
        :returns: self.
        :raises IllegalArgumentException: raises the exception if parameters are
            not expected type.
        """
        CheckValue.check_not_none(value, 'value')
        CheckValue.check_boolean(is_inclusive, 'is_inclusive')
        self._end = value
        self._end_inclusive = is_inclusive
        return self

    def get_end(self):
        """
        Returns the field value that defines upper bound of the range, or None
        if no upper bound is enforced.

        :returns: the end field value.
        """
        return self._end

    def get_end_inclusive(self):
        """
        Returns whether end is included in the range, i.e., end is greater than
        or equal to the last field value in the range. This value is valid only
        if the end value is not None.

        :returns: True if the end value is inclusive.
        :rtype: bool
        """
        return self._end_inclusive

    def validate(self):
        # Ensures that the object is self-consistent and if not, throws
        # IllegalArgumentException. Validation of the range values themselves is
        # done remotely.
        start_type = None if self._start is None else type(self._start)
        end_type = None if self._end is None else type(self._end)
        if start_type is None and end_type is None:
            raise IllegalArgumentException(
                'FieldRange: must specify a start or end value.')
        if (start_type is not None and end_type is not None and
                start_type is not end_type):
            raise IllegalArgumentException(
                'FieldRange: Mismatch of start and end types. Start type is ' +
                str(start_type) + ', end type is ' + str(end_type))


class HttpConstants(object):
    # The name of the Authorization header.
    AUTHORIZATION = 'Authorization'

    # The name of the date header.
    DATE = 'date'

    # The current version of the protocol
    NOSQL_VERSION = 'V2'

    # The name of the opc obo token header.
    OPC_OBO_TOKEN = 'opc-obo-token'

    # A header for transferring the compartment id on an http request.
    REQUEST_COMPARTMENT_ID = 'x-nosql-compartment-id'

    # The http header that identifies the client scoped unique request id
    # associated with each request. The request header is returned by the
    # server, as part of the response and serves to associate the response with
    # the request.
    #
    # Note: We could use stream ids to associate a request and response.
    # However, the current handler pipeline on the client side operates at the
    # http level rather than the frame level, and consequently does not have
    # access to the stream id.
    REQUEST_ID_HEADER = 'x-nosql-request-id'

    # The name of the (request-target) header.
    REQUEST_TARGET = '(request-target)'

    # Creates a URI path from the arguments
    def _make_path(*args):
        path = args[0]
        for index in range(1, len(args)):
            path += '/' + args[index]
        return path

    # The base path to the on-premise security services. All users need a
    # leading "/" so add it here.
    KV_SECURITY_PATH = _make_path('/' + NOSQL_VERSION, 'nosql/security')

    # The path denoting a NoSQL request
    NOSQL_DATA_PATH = _make_path(NOSQL_VERSION, 'nosql/data')


class IndexInfo(object):
    """
    IndexInfo represents the information about a single index including its name
    and field names. Instances of this class are returned in
    :py:class:`GetIndexesResult`.
    """

    def __init__(self, index_name, field_names):
        self._index_name = index_name
        self._field_names = field_names

    def __str__(self):
        return ('IndexInfo [indexName=' + self._index_name + ', fields=[' +
                ','.join(self._field_names) + ']]')

    def get_index_name(self):
        """
        Returns the name of the index.

        :returns: the index name.
        :rtype: str
        """
        return self._index_name

    def get_field_names(self):
        """
        Returns the list of field names that define the index.

        :returns: the field names.
        :rtype: list(str)
        """
        return self._field_names


class JsonNone(object):
    """
    Represents an JSON NONE field value.
    """

    def __str__(self):
        return 'JsonNone'


class LogUtils(object):

    # Utility methods to facilitate Logging.
    def __init__(self, logger=None):
        self._logger = logger

    def log_critical(self, msg):
        if self._logger is not None:
            self._logger.critical(msg)

    def log_error(self, msg):
        if self._logger is not None:
            self._logger.error(msg)

    def log_warning(self, msg):
        if self._logger is not None:
            self._logger.warning(msg)

    def log_info(self, msg):
        if self._logger is not None:
            self._logger.info(msg)

    def log_debug(self, msg):
        if self._logger is not None:
            self._logger.debug(msg)

    def is_enabled_for(self, level):
        return self._logger is not None and self._logger.isEnabledFor(level)

    def set_level(self, level):
        return self._logger is not None and self._logger.setLevel(level)

    def get_logger(self):
        return self._logger


class Memoize(object):

    # A cache that used for saving the access token.
    def __init__(self, duration=60):
        self._cache = {}
        self._duration = duration
        self.lock = Lock()

    @synchronized
    def set(self, key, value):
        self._cache[key] = {'value': value, 'time': time()}

    @synchronized
    def get(self, key):
        if key in self._cache and not self._is_obsolete(self._cache[key]):
            return self._cache[key]['value']
        return None

    def _is_obsolete(self, entry):
        return time() - entry['time'] > self._duration


class PackedInteger(object):
    # The maximum number of bytes needed to store an int value (5).
    MAX_LENGTH = 5

    # The maximum number of bytes needed to store a long value (9).
    MAX_LONG_LENGTH = 9

    @staticmethod
    def write_sorted_int(buf, offset, value):
        """
        Writes a packed sorted integer starting at the given buffer offset and
        returns the next offset to be written.

        :param buf: the buffer to write to.
        :type buf: bytearray
        :param offset: the offset in the buffer at which to start writing.
        :type offset: int
        :param value: the integer to be written.
        :type value: int
        :returns: the offset past the bytes written.
        :rtype: int

        Values in the inclusive range [-119,120] are stored in a single byte.
        For values outside that range, the first byte stores the number of
        additional bytes. The additional bytes store (value + 119 for negative
        and value - 121 for positive) as an unsigned big endian integer.
        """
        b1 = offset
        offset += 1
        if value < -119:
            """
            If the value < -119, then first adjust the value by adding 119. Then
            the adjusted value is stored as an unsigned big endian integer.
            """
            value += 119

            """
            Store the adjusted value as an unsigned big endian integer. For an
            negative integer, from left to right, the first significant byte is
            the byte which is not equal to 0xFF. Also please note that, because
            the adjusted value is stored in big endian integer, we extract the
            significant byte from left to right.

            In the left to right order, if the first byte of the adjusted value
            is a significant byte, it will be stored in the 2nd byte of the buf.
            Then we will look at the 2nd byte of the adjusted value to see if
            this byte is the significant byte, if yes, this byte will be stored
            in the 3rd byte of the buf, and the like.
            """
            if value | 0x00FFFFFF != 0xFFFFFFFF:
                buf[offset] = value >> 24 & 0xFF
                offset += 1
            if value | 0x0000FFFF != 0xFFFFFFFF:
                buf[offset] = value >> 16 & 0xFF
                offset += 1
            if value | 0x000000FF != 0xFFFFFFFF:
                buf[offset] = value >> 8 & 0xFF
                offset += 1
            buf[offset] = value & 0xFF
            offset += 1

            """
            value_len is the length of the value part stored in buf. Because
            the first byte of buf is used to stored the length, we need to
            subtract one.
            """
            value_len = offset - b1 - 1

            """
            The first byte stores the number of additional bytes. Here we store
            the result of 0x08 - value_len, rather than directly store
            value_len. The reason is to implement natural sort order for
            byte-by-byte comparison.
            """
            buf[b1] = (0x08 - value_len) & 0xFF
        elif value > 120:
            """
            If the value > 120, then first adjust the value by subtracting 121.
            Then the adjusted value is stored as an unsigned big endian integer.
            """
            value -= 121

            """
            Store the adjusted value as an unsigned big endian integer. For a
            positive integer, from left to right, the first significant byte is
            the byte which is not equal to 0x00.

            In the left to right order, if the first byte of the adjusted value
            is a significant byte, it will be stored in the 2nd byte of the buf.
            Then we will look at the 2nd byte of the adjusted value to see if
            this byte is the significant byte, if yes, this byte will be stored
            in the 3rd byte of the buf, and the like.
            """
            if value & 0xFF000000 != 0:
                buf[offset] = value >> 24 & 0xFF
                offset += 1
            if value & 0xFFFF0000 != 0:
                buf[offset] = value >> 16 & 0xFF
                offset += 1
            if value & 0xFFFFFF00 != 0:
                buf[offset] = value >> 8 & 0xFF
                offset += 1
            buf[offset] = value & 0xFF
            offset += 1

            """
            value_len is the length of the value part stored in buf. Because the
            first byte of buf is used to stored the length, we need to subtract
            one.
            """
            value_len = offset - b1 - 1

            """
            The first byte stores the number of additional bytes. Here we store
            the result of 0xF7 + value_len, rather than directly store
            value_len. The reason is to implement natural sort order for
            byte-by-byte comparison.
            """
            buf[b1] = (0xF7 + value_len) & 0xFF
        else:
            """
            If -119 <= value <= 120, only one byte is needed to store the value.
            The stored value is the original value plus 127.
            """
            buf[b1] = (value + 127) & 0xFF
        return offset

    @staticmethod
    def write_sorted_long(buf, offset, value):
        """
        Writes a packed sorted long integer starting at the given buffer offset
        and returns the next offset to be written.

        :param buf: the buffer to write to.
        :type buf: bytearray
        :param offset: the offset in the buffer at which to start writing.
        :type offset: int
        :param value: the long integer to be written.
        :type value: int for python 3 and long for python 2
        :returns: the offset past the bytes written.
        :rtype: int

        Values in the inclusive range [-119,120] are stored in a single byte.
        For values outside that range, the first byte stores the number of
        additional bytes. The additional bytes store (value + 119 for negative
        and value - 121 for positive) as an unsigned big endian integer.
        """
        b1 = offset
        offset += 1
        if value < -119:
            """
            If the value < -119, then first adjust the value by adding 119. Then
            the adjusted value is stored as an unsigned big endian integer.
            """
            value += 119

            """
            Store the adjusted value as an unsigned big endian integer. For an
            negative integer, from left to right, the first significant byte is
            the byte which is not equal to 0xFF. Also please note that, because
            the adjusted value is stored in big endian integer, we extract the
            significant byte from left to right.

            In the left to right order, if the first byte of the adjusted value
            is a significant byte, it will be stored in the 2nd byte of the buf.
            Then we will look at the 2nd byte of the adjusted value to see if
            this byte is the significant byte, if yes, this byte will be stored
            in the 3rd byte of the buf, and the like.
            """
            if value | 0x00FFFFFFFFFFFFFF != 0xFFFFFFFFFFFFFFFF:
                buf[offset] = value >> 56 & 0xFF
                offset += 1
            if value | 0x0000FFFFFFFFFFFF != 0xFFFFFFFFFFFFFFFF:
                buf[offset] = value >> 48 & 0xFF
                offset += 1
            if value | 0x000000FFFFFFFFFF != 0xFFFFFFFFFFFFFFFF:
                buf[offset] = value >> 40 & 0xFF
                offset += 1
            if value | 0x00000000FFFFFFFF != 0xFFFFFFFFFFFFFFFF:
                buf[offset] = value >> 32 & 0xFF
                offset += 1
            if value | 0x0000000000FFFFFF != 0xFFFFFFFFFFFFFFFF:
                buf[offset] = value >> 24 & 0xFF
                offset += 1
            if value | 0x000000000000FFFF != 0xFFFFFFFFFFFFFFFF:
                buf[offset] = value >> 16 & 0xFF
                offset += 1
            if value | 0x00000000000000FF != 0xFFFFFFFFFFFFFFFF:
                buf[offset] = value >> 8 & 0xFF
                offset += 1
            buf[offset] = value & 0xFF
            offset += 1

            """
            value_len is the length of the value part stored in buf. Because the
            first byte of buf is used to stored the length, so we need to minus
            one.
            """
            value_len = offset - b1 - 1

            """
            The first byte stores the number of additional bytes. Here we store
            the result of 0x08 - value_len, rather than directly store
            value_len. The reason is to implement nature sort order for
            byte-by-byte comparison.
            """
            buf[b1] = (0x08 - value_len) & 0xFF
        elif value > 120:
            """
            If the value > 120, then first adjust the value by subtracting 119.
            Then the adjusted value is stored as an unsigned big endian integer.
            """
            value -= 121

            """
            Store the adjusted value as an unsigned big endian integer. For a
            positive integer, from left to right, the first significant byte is
            the byte which is not equal to 0x00.

            In the left to right order, if the first byte of the adjusted value
            is a significant byte, it will be stored in the 2nd byte of the buf.
            Then we will look at the 2nd byte of the adjusted value to see if
            this byte is the significant byte, if yes, this byte will be stored
            in the 3rd byte of the buf, and the like.
            """
            if value & 0xFF00000000000000 != 0:
                buf[offset] = value >> 56 & 0xFF
                offset += 1
            if value & 0xFFFF000000000000 != 0:
                buf[offset] = value >> 48 & 0xFF
                offset += 1
            if value & 0xFFFFFF0000000000 != 0:
                buf[offset] = value >> 40 & 0xFF
                offset += 1
            if value & 0xFFFFFFFF00000000 != 0:
                buf[offset] = value >> 32 & 0xFF
                offset += 1
            if value & 0xFFFFFFFFFF000000 != 0:
                buf[offset] = value >> 24 & 0xFF
                offset += 1
            if value & 0xFFFFFFFFFFFF0000 != 0:
                buf[offset] = value >> 16 & 0xFF
                offset += 1
            if value & 0xFFFFFFFFFFFFFF00 != 0:
                buf[offset] = value >> 8 & 0xFF
                offset += 1
            buf[offset] = value & 0xFF
            offset += 1

            """
            value_en is the length of the value part stored in buf. Because the
            first byte of buf is used to stored the length, so we need to minus
            one.
            """
            value_len = offset - b1 - 1

            """
            The first byte stores the number of additional bytes. Here we store
            the result of 0xF7 + value_len, rather than directly store
            value_len. The reason is to implement nature sort order for
            byte-by-byte comparison.
            """
            buf[b1] = (0xF7 + value_len) & 0xFF
        else:
            """
            If -119 <= value <= 120, only one byte is needed to store the value.
            The stored value is the original value adds 127.
            """
            buf[b1] = (value + 127) & 0xFF
        return offset

    @staticmethod
    def get_read_sorted_int_length(buf, offset):
        """
        Returns the number of bytes that would be read by
        :py:meth:`read_sorted_int`.

        Because the length is stored in the first byte, this method may be
        called with only the first byte of the packed integer in the given
        buffer. This method only accesses one byte at the given offset.

        :param buf: the buffer to read from.
        :type buf: bytearray
        :param offset: the offset in the buffer at which to start reading.
        :type offset: int
        :returns: the number of bytes that would be read.
        :rtype: int
        """
        # The first byte of the buf stores the length of the value part.
        b1 = buf[offset] & 0xff
        if b1 < 0x08:
            return 1 + 0x08 - b1
        if b1 > 0xf7:
            return 1 + b1 - 0xf7
        return 1

    @staticmethod
    def get_read_sorted_long_length(buf, offset):
        """
        Returns the number of bytes that would be read by
        :py:meth:`read_sorted_long`.

        Because the length is stored in the first byte, this method may be
        called with only the first byte of the packed integer in the given
        buffer. This method only accesses one byte at the given offset.

        :param buf: the buffer to read from.
        :type buf: bytearray
        :param offset: the offset in the buffer at which to start reading.
        :type offset: int
        :returns: the number of bytes that would be read.
        :rtype: int
        """
        # The length is stored in the same way for int and long.
        return PackedInteger.get_read_sorted_int_length(buf, offset)

    @staticmethod
    def read_sorted_int(buf, offset):
        """
        Reads a sorted packed integer at the given buffer offset and returns it.

        :param buf: the buffer to read from.
        :type buf: bytearray
        :param offset: the offset in the buffer at which to start reading.
        :type offset: int
        :returns: the integer that was read.
        :rtype: int
        """
        # The first byte of the buf stores the length of the value part.
        b1 = buf[offset] & 0xff
        offset += 1
        # Adjust the byte_len to the real length of the value part.
        if b1 < 0x08:
            byte_len = 0x08 - b1
            negative = True
        elif b1 > 0xf7:
            byte_len = b1 - 0xf7
            negative = False
        else:
            return b1 - 127

        """
        The following bytes on the buf store the value as a big endian integer.
        We extract the significant bytes from the buf and put them into the
        value in big endian order.
        """
        if negative:
            value = -1
        else:
            value = 0
        if byte_len > 3:
            value = (value << 8) | (buf[offset] & 0xFF)
            offset += 1
        if byte_len > 2:
            value = (value << 8) | (buf[offset] & 0xFF)
            offset += 1
        if byte_len > 1:
            value = (value << 8) | (buf[offset] & 0xFF)
            offset += 1
        value = (value << 8) | (buf[offset] & 0xFF)
        offset += 1

        """
        After get the adjusted value, we have to adjust it back to the original
        value.
        """
        if negative:
            value -= 119
        else:
            value += 121
        return value

    @staticmethod
    def read_sorted_long(buf, offset):
        """
        Reads a sorted packed long integer at the given buffer offset and
        returns it.

        :param buf: the buffer to read from.
        :type buf: bytearray
        :param offset: the offset in the buffer at which to start reading.
        :type offset: int
        :returns: the long integer that was read.
        :rtype: int for python 3 and long for python 2
        """
        # The first byte of the buf stores the length of the value part.
        b1 = buf[offset] & 0xff
        offset += 1
        # Adjust the byte_len to the real length of the value part.
        if b1 < 0x08:
            byte_len = 0x08 - b1
            negative = True
        elif b1 > 0xf7:
            byte_len = b1 - 0xf7
            negative = False
        else:
            try:
                return long(b1 - 127)
            except NameError:
                return b1 - 127

        """
        The following bytes on the buf store the value as a big endian integer.
        We extract the significant bytes from the buf and put them into the
        value in big endian order.
        """
        if negative:
            value = -1
        else:
            value = 0
        if byte_len > 7:
            value = (value << 8) | (buf[offset] & 0xFF)
            offset += 1
        if byte_len > 6:
            value = (value << 8) | (buf[offset] & 0xFF)
            offset += 1
        if byte_len > 5:
            value = (value << 8) | (buf[offset] & 0xFF)
            offset += 1
        if byte_len > 4:
            value = (value << 8) | (buf[offset] & 0xFF)
            offset += 1
        if byte_len > 3:
            value = (value << 8) | (buf[offset] & 0xFF)
            offset += 1
        if byte_len > 2:
            value = (value << 8) | (buf[offset] & 0xFF)
            offset += 1
        if byte_len > 1:
            value = (value << 8) | (buf[offset] & 0xFF)
            offset += 1
        value = (value << 8) | (buf[offset] & 0xFF)
        offset += 1
        """
        After obtaining the adjusted value, we have to adjust it back to the
        original value.
        """
        if negative:
            value -= 119
        else:
            value += 121
        try:
            return long(value)
        except NameError:
            return value


class PreparedStatement(object):
    """
    A class encapsulating a prepared query statement. It includes state that can
    be sent to a server and executed without re-parsing the query. It includes
    bind variables which may be set for each successive use of the query. The
    prepared query itself is read-only but this object contains a dictionary of
    bind variables and is not thread-safe if variables are used.

    PreparedStatement instances are returned inside :py:class:`PrepareResult`
    objects returned by :py:meth:`NoSQLHandle.prepare`

    A single instance of PreparedStatement is thread-safe if bind variables are
    not used. If bind variables are to be used and the statement shared among
    threads additional instances of PreparedStatement can be constructed using
    :py:meth:`copy_statement`.
    """
    OPCODE_SELECT = 5

    def __init__(self, sql_text, query_plan, topology_info, proxy_statement,
                 driver_plan, num_iterators, num_registers, external_vars,
                 namespace, table_name, operation):
        """
        Constructs a PreparedStatement. Construction is hidden to eliminate
        application access to the underlying statement, reducing the chance of
        corruption.
        """
        # 10 is arbitrary. TODO: put magic number in it for validation?
        if proxy_statement is None or len(proxy_statement) < 10:
            raise IllegalArgumentException(
                'Invalid prepared query, cannot be None.')

        self._sql_text = sql_text
        self._query_plan = query_plan
        # Applicable to advanced queries only.
        self._topology_info = topology_info
        # The serialized PreparedStatement created at the backend store. It is
        # opaque for the driver. It is received from the proxy and sent back to
        # the proxy every time a new batch of results is needed.
        self._proxy_statement = proxy_statement
        # The part of the query plan that must be executed at the driver. It is
        # received from the proxy when the query is prepared there. It is
        # deserialized by the driver and not sent back to the proxy again.
        # Applicable to advanced queries only.
        self._driver_query_plan = driver_plan
        # The number of iterators in the full query plan
        # Applicable to advanced queries only.
        self._num_iterators = num_iterators
        # The number of registers required to run the full query plan.
        # Applicable to advanced queries only.
        self._num_registers = num_registers
        # Maps the name of each external variable to its id, which is a position
        # in a field value array stored in the RuntimeControlBlock and holding
        # the values of the variables. Applicable to advanced queries only.
        self._variables = external_vars
        # The values for the external variables of the query. This dict is
        # populated by the application. It is sent to the proxy every time a new
        # batch of results is needed. The values in this dict are also placed in
        # the RuntimeControlBlock field value array, just before the query
        # starts its execution at the driver.
        self._bound_variables = None
        self._namespace = namespace
        self._table_name = table_name
        self._operation = operation
        self.lock = Lock()

    def clear_variables(self):
        """
        Clears all bind variables from the statement.
        """
        if self._bound_variables is not None:
            self._bound_variables.clear()

    def copy_statement(self):
        """
        Returns a new instance that shares this object's prepared query, which
        is immutable, but does not share its variables.

        :returns: a new PreparedStatement using this instance's prepared query.
            Bind variables are uninitialized.
        :rtype: PreparedStatement
        """
        return PreparedStatement(
            self._sql_text, self._query_plan, self._topology_info,
            self._proxy_statement, self._driver_query_plan, self._num_iterators,
            self._num_registers, self._variables, self._namespace,
            self._table_name, self._operation)

    def does_writes(self):
        # if it's not SELECT, it does writes.
        return self._operation != PreparedStatement.OPCODE_SELECT

    def driver_plan(self):
        return self._driver_query_plan

    def get_namespace(self):
        # Return namespace from prepared statement, if any.
        return self._namespace

    def get_query_plan(self):
        """
        Returns a string representation of the query execution plan, if it was
        requested in the :py:class:`PrepareRequest`; None otherwise.

        :returns: the string representation of the query execution plan.
        :rtype: bool
        """
        return self._query_plan

    def get_sql_text(self):
        """
        Returns the SQL text of this PreparedStatement.

        :returns: the SQL text of this PreparedStatement.
        :rtype: str
        """
        return self._sql_text

    def get_statement(self):
        # internal use to return the serialized, prepared query, opaque
        return self._proxy_statement

    def get_table_name(self):
        # Return table name from prepared statement, if any.
        return self._table_name

    def get_variables(self):
        """
        Returns the dictionary of variables to use for a prepared query
        with variables.

        :returns: the dictionary.
        :rtype: dict
        """
        return self._bound_variables

    def get_variable_values(self):
        if self._bound_variables is None:
            return None
        values = [0] * len(self._bound_variables)
        for key in self._bound_variables:
            varid = self._variables.get(key)
            values[varid] = self._bound_variables[key]
        return values

    def is_simple_query(self):
        return self._driver_query_plan is None

    def num_iterators(self):
        return self._num_iterators

    def num_registers(self):
        return self._num_registers

    def print_driver_plan(self):
        if self._driver_query_plan is not None:
            return self._driver_query_plan.display()
        return None

    @synchronized
    def set_topology_info(self, topology_info):
        if topology_info is None:
            return
        if self._topology_info is None:
            self._topology_info = topology_info
            return
        if self._topology_info.get_seq_num() < topology_info.get_seq_num():
            self._topology_info = topology_info

    def set_variable(self, variable, value):
        """
        Binds an external variable to a given value. The variable is identified
        by its name or its position within the query string. The variable that
        appears first in the query text has position 1, the variable that
        appears second has position 2 and so on.

        :param variable: the name or the position of the variable.
        :type variable: str or int
        :param value: the value.
        :type value: a value matching the type of the field
        :returns: self.
        :raises IllegalArgumentException: raises the exception if variable is
            not a string or positive integer.
        """
        if (not (CheckValue.is_str(variable) or CheckValue.is_int(variable)) or
                CheckValue.is_int(variable) and variable <= 0):
            raise IllegalArgumentException(
                'variable must be a string or positive integer.')
        if isinstance(variable, str):
            if self._bound_variables is None:
                self._bound_variables = dict()
            if (self._variables is not None and
                    self._variables.get(variable) is None):
                raise IllegalArgumentException(
                    'The query does not contain the variable: ' + variable)
            self._bound_variables[variable] = value
            return self
        else:
            if self._variables is None:
                name = '#' + str(variable)
                return self.set_variable(name, value)
            search_id = variable - 1
            for (k, v) in self._variables.items():
                if v == search_id:
                    return self.set_variable(k, value)
            raise IllegalArgumentException(
                'There is no external variable at position ' + str(variable))

    def topology_info(self):
        return self._topology_info

    @synchronized
    def topology_seq_num(self):
        return (-1 if self._topology_info is None else
                self._topology_info.get_seq_num())


class PutOption(object):
    """
    Set the put option for put requests.
    """
    IF_ABSENT = 0
    """Set PutOption.IF_ABSENT to perform put if absent operation."""
    IF_PRESENT = 1
    """Set PutOption.IF_PRESENT to perform put if present operation."""
    IF_VERSION = 2
    """Set PutOption.IF_VERSION to perform put if version operation."""


class ResourcePrincipalClaimKeys(object):
    """
    Claim keys in the resource principal session token(RPST).

    They can be used to retrieve resource principal metadata such as its
    compartment and tenancy OCID.

    """
    COMPARTMENT_ID_CLAIM_KEY = 'res_compartment'
    """
    The claim name that the RPST holds for the resource compartment. This can be
    passed to
    :py:meth:`borneo.iam.SignatureProvider.get_resource_principal_claim` to
    retrieve the resource's compartment OCID.
    """
    TENANT_ID_CLAIM_KEY = 'res_tenant'
    """
    The claim name that the RPST holds for the resource tenancy. This can be
    passed to
    :py:meth:`borneo.iam.SignatureProvider.get_resource_principal_claim` to
    retrieve the resource's tenancy OCID.
    """


class SSLAdapter(adapters.HTTPAdapter):
    """
    Internal use only.

    An HTTPS Transport Adapter that uses an arbitrary SSLContext.
    """

    def __init__(self, ssl_context=None, **kwargs):
        self._ssl_context = ssl_context
        super(SSLAdapter, self).__init__(**kwargs)

    def init_poolmanager(self, *args, **kwargs):
        kwargs['ssl_context'] = self._ssl_context
        return super(SSLAdapter, self).init_poolmanager(*args, **kwargs)

    def proxy_manager_for(self, *args, **kwargs):
        kwargs['ssl_context'] = self._ssl_context
        return super(SSLAdapter, self).proxy_manager_for(*args, **kwargs)


class State(object):
    """
    Represents the table state.
    """
    ACTIVE = 'ACTIVE'
    """Represents the table is active."""
    CREATING = 'CREATING'
    """Represents the table is creating."""
    DROPPED = 'DROPPED'
    """Represents the table is dropped."""
    DROPPING = 'DROPPING'
    """Represents the table is dropping."""
    UPDATING = 'UPDATING'
    """Represents the table is updating."""


class SystemState(object):
    """
    On-premise only.

    The current state of the system request.
    """
    COMPLETE = 'COMPLETE'
    """
    The operation is complete and was successful. Failures are thrown as
    exceptions.
    """
    WORKING = 'WORKING'
    """The operation is in progress."""


class Durability(object):
    """
    Durability defines the durability characteristics associated with a standalone write
    (put or update) operation.

    This is currently only supported in On-Prem installations. It is ignored
    in the cloud service.

    The overall durability is a function of the SYNC_POLICY and
    ACK_POLICY in effect for the Master, and the SYNC_POLICY in
    effect for each Replica.

    SYNC_POLICY represents policies to be used when committing a
    transaction. High levels of synchronization offer a greater guarantee
    that the transaction is persistent to disk, but trade that off for
    lower performance. The possible SYNC_POLICY values are:

    * SYNC writes and synchronously flushes the log on transaction commit.
      Transactions exhibit all the ACID (atomicity, consistency,
      isolation, and durability) properties.

    * NO_SYNC does not write or synchronously flush the log on transaction
      commit. Transactions exhibit the ACI (atomicity, consistency, and
      isolation) properties, but not D (durability); that is, database
      integrity will be maintained, but if the application or system fails,
      it is possible some number of the most recently committed transactions
      may be undone during recovery. The number of transactions at risk is
      governed by how many log updates can fit into the log buffer, how
      often the operating system flushes dirty buffers to disk, and how
      often log checkpoints occur.

    * WRITE_NO_SYNC writes but does not synchronously flush the log on
      transaction commit. Transactions exhibit the ACI (atomicity, consistency,
      and isolation) properties, but not D (durability); that is, database
      integrity will be maintained, but if the operating system fails, it is
      possible some number of the most recently committed transactions may be
      undone during recovery. The number of transactions at risk is
      governed by how often the operating system flushes dirty buffers to
      disk, and how often log checkpoints occur.

    REPLICA_ACK_POLICY defines the policy for how replicated commits are handled.
    A replicated environment makes it possible to increase an application's
    transaction commit guarantees by committing changes to its replicas on
    the network.

    Possible REPLICA_ACK_POLICY values include:

    * ALL defines that all replicas must acknowledge that they
      have committed the transaction. This policy should be selected only if
      your replication group has a small number of replicas, and those
      replicas are on extremely reliable networks and servers.

    * NONE defines that no transaction commit acknowledgments
      are required and the master will never wait for replica acknowledgments.
      In this case, transaction durability is determined entirely by the type
      of commit that is being performed on the master.

    * SIMPLE_MAJORITY defines that a simple majority of replicas
      must acknowledge that they have committed the transaction. This
      acknowledgment policy, in conjunction with an election policy which
      requires at least a simple majority, ensures that the changes made by
      the transaction remains durable if a new election is held.

    The default Durability is configured in the proxy server with which this
    SDK communicates. It is an optional startup parameter.
    """

    SYNC_POLICY = enum(SYNC=1,
                       NO_SYNC=2,
                       WRITE_NO_SYNC=3)
    """
    SYNC_POLICY
    """

    REPLICA_ACK_POLICY = enum(ALL=1,
                              NONE=2,
                              SIMPLE_MAJORITY=2)
    """
    REPLICA_ACK_POLICY
    """

    def __init__(self, master_sync, replica_sync, replica_ack):
        """
        Create a Durability object

        :param master_sync: the master sync policy
        :type master_sync: SYNC_POLICY
        :param replica_sync: the replica sync policy
        :type replica_sync: SYNC_POLICY
        :param replica_ack: the replica ack policy
        :type replica_ack: REPLICA_ACK_POLICY
        """
        self.master_sync = master_sync
        self.replica_sync = replica_sync
        self.replica_ack = replica_ack

    # noinspection PyUnresolvedReferences
    def validate(self):
        # an all-zero Durability specifies "use the durability settings on the server."
        if self.master_sync == 0 and self.replica_sync == 0 and self.replica_ack == 0:
            return
        if (self.master_sync != Durability.SYNC_POLICY.SYNC and
                self.master_sync != Durability.SYNC_POLICY.NO_SYNC and
                self.master_sync != Durability.SYNC_POLICY.WRITE_NO_SYNC):
            raise IllegalArgumentException(
                'Invalid value for Durability master_sync.')
        if (self.replica_sync != Durability.SYNC_POLICY.SYNC and
                self.replica_sync != Durability.SYNC_POLICY.NO_SYNC and
                self.replica_sync != Durability.SYNC_POLICY.WRITE_NO_SYNC):
            raise IllegalArgumentException(
                'Invalid value for Durability replica_sync.')
        if (self.replica_ack != Durability.REPLICA_ACK_POLICY.ALL and
                self.replica_ack != Durability.REPLICA_ACK_POLICY.NONE and
                self.replica_ack != Durability.REPLICA_ACK_POLICY.SIMPLE_MAJORITY):
            raise IllegalArgumentException(
                'Invalid value for Durability replica_ack.')


class TableLimits(object):
    """
    Cloud service only.

    A TableLimits instance is used during table creation to specify the
    throughput and capacity to be consumed by the table. It is also used in an
    operation to change the limits of an existing table.
    :py:meth:`NoSQLHandle.table_request` and :py:class:`TableRequest` are used
    to perform these operations. These values are enforced by the system and
    used for billing purposes.

    Throughput limits are defined in terms of read units and write units. A read
    unit represents 1 eventually consistent read per second for data up to 1 KB
    in size. A read that is absolutely consistent is double that, consuming 2
    read units for a read of up to 1 KB in size. This means that if an
    application is to use Consistency.ABSOLUTE it may need to specify additional
    read units when creating a table. A write unit represents 1 write per second
    of data up to 1 KB in size.

    In addition to throughput table capacity must be specified to indicate the
    maximum amount of storage, in gigabytes, allowed for the table.

    In provisioned mode, all 3 values must be used whenever using this object.
    There are no defaults and no mechanism to indicate "no change."

    In on demand mode, only the storage_gb parameter must be set.

    :param read_units: the desired throughput of read operation in terms of read
        units. A read unit represents 1 eventually consistent read per second
        for data up to 1 KB in size. A read that is absolutely consistent is
        double that, consuming 2 read units for a read of up to 1 KB in size.
    :type read_units: int
    :param write_units: the desired throughput of write operation in terms of
        write units. A write unit represents 1 write per second of data up to 1
        KB in size.
    :type write_units: int
    :param storage_gb: the maximum storage to be consumed by the table, in
        gigabytes.
    :type storage_gb: int
    :param mode: the mode of the table: provisioned (the default) or on demand.
    :type mode: CAPACITY_MODE
    :raises IllegalArgumentException: raises the exception if parameters are not
        valid.
    :versionchanged: 5.3.0, added optional CAPACITY_MODE
    """

    """
    TableLimits includes an optional mode

    :versionadded: 5.3.0
    """
    CAPACITY_MODE = enum(PROVISIONED=1,
                         ON_DEMAND=2)

    # noinspection PyUnresolvedReferences
    def __init__(self, read_units, write_units, storage_gb,
                 mode=CAPACITY_MODE.PROVISIONED):
        """
        Creates a TableLimits object

        :param read_units: the desired throughput of read operation in terms
         of read units. A read unit represents 1 eventually consistent read
         per second for data up to 1 KB in size. A read that is absolutely
         consistent is double that, consuming 2 read units for a read of up to
         1 KB in size.
        :type read_units: int
        :param write_units: the desired throughput of write operation in terms
         of write units. A write unit represents 1 write per second of data up
         to 1 KB in size.
        :type write_units: int
        :param storage_gb: the maximum storage to be consumed by the table, in
         gigabytes.
        :type storage_gb: int
        :param mode: the mode of the table: provisioned (the default) or on
         demand.
        :type mode: CAPACITY_MODE
        :raises IllegalArgumentException: raises the exception if parameters
         are not valid.
        :versionchanged: 5.3.0, added optional CAPACITY_MODE
       """

        # Constructs a TableLimits instance.
        self._mode = None
        CheckValue.check_int(read_units, 'read_units')
        CheckValue.check_int(write_units, 'write_units')
        CheckValue.check_int(storage_gb, 'storage_gb')
        self._read_units = read_units
        self._write_units = write_units
        self._storage_gb = storage_gb
        self.set_mode(mode)

    def __str__(self):
        return ('[' + str(self._read_units) + ', ' + str(self._write_units) +
                ', ' + str(self._storage_gb) + ']')

    def set_read_units(self, read_units):
        """
        Sets the read throughput in terms of read units.

        :param read_units: the throughput to use, in read units.
        :type read_units: int
        :returns: self.
        :raises IllegalArgumentException: raises the exception if read_units is
            not a integer.
        """
        CheckValue.check_int(read_units, 'read_units')
        self._read_units = read_units
        return self

    def get_read_units(self):
        """
        Returns the read throughput in terms of read units.

        :returns: the read units.
        :rtype: int
        """
        return self._read_units

    def set_write_units(self, write_units):
        """
        Sets the write throughput in terms of write units.

        :param write_units: the throughput to use, in write units.
        :type write_units: int
        :returns: self.
        :raises IllegalArgumentException: raises the exception if write_units is
            not a integer.
        """
        CheckValue.check_int(write_units, 'write_units')
        self._write_units = write_units
        return self

    def get_write_units(self):
        """
        Returns the write throughput in terms of write units.

        :returns: the write units.
        :rtype: int
        """
        return self._write_units

    def set_storage_gb(self, storage_gb):
        """
        Sets the storage capacity in gigabytes.

        :param storage_gb: the capacity to use, in gigabytes.
        :type storage_gb: int
        :returns: self.
        :raises IllegalArgumentException: raises the exception if storage_gb is
            not a integer.
        """
        CheckValue.check_int(storage_gb, 'storage_gb')
        self._storage_gb = storage_gb
        return self

    def get_storage_gb(self):
        """
        Returns the storage capacity in gigabytes.

        :returns: the storage capacity in gigabytes.
        :rtype: int
        """
        return self._storage_gb

    # noinspection PyUnresolvedReferences
    def set_mode(self, mode):
        """
        Sets the mode of the table:
            PROVISIONED: Fixed maximum read/write units. This is the default.
            ON_DEMAND: Flexible read/write limits.

        :param mode: the capacity to use, in gigabytes.
        :type mode: TableLimits.CAPACITY_MODE
        :returns: self.
        :raises IllegalArgumentException: raises the exception if mode is
            invalid.
        :versionadded: 5.3.0
        """
        if (mode != TableLimits.CAPACITY_MODE.PROVISIONED and
                mode != TableLimits.CAPACITY_MODE.ON_DEMAND):
            raise IllegalArgumentException(
                'TableLimits mode must be one of PROVISIONED or ON_DEMAND')
        self._mode = mode

    def get_mode(self):
        """
        Returns the capacity mode of the table.

        :returns: mode: PROVISIONED or ON_DEMAND
        :versionadded: 5.3.0
        """
        return self._mode

    # noinspection PyUnresolvedReferences
    def validate(self):
        if self._storage_gb <= 0:
            raise IllegalArgumentException(
                'TableLimits values must be non-negative.')
        if (self._mode != TableLimits.CAPACITY_MODE.ON_DEMAND and
                (self._read_units <= 0 or self._write_units <= 0)):
            raise IllegalArgumentException(
                'TableLimits values must be non-negative.')


class TableUsage(object):
    """
    TableUsage represents a single usage record, or slice, that includes
    information about read and write throughput consumed during that period as
    well as the current information regarding storage capacity. In addition the
    count of throttling exceptions for the period is reported.
    """

    def __init__(self, start_time_ms, seconds_in_period, read_units,
                 write_units, storage_gb, read_throttle_count,
                 write_throttle_count, storage_throttle_count):
        # Internal use only.
        self._start_time_ms = start_time_ms
        self._seconds_in_period = seconds_in_period
        self._read_units = read_units
        self._write_units = write_units
        self._storage_gb = storage_gb
        self._read_throttle_count = read_throttle_count
        self._write_throttle_count = write_throttle_count
        self._storage_throttle_count = storage_throttle_count

    def __str__(self):
        return ('TableUsage [start_time_ms=' + str(self._start_time_ms) +
                ', seconds_in_period=' + str(self._seconds_in_period) +
                ', read_units=' + str(self._read_units) + ', write_units=' +
                str(self._write_units) + ', storage_gb=' +
                str(self._storage_gb) + ', read_throttle_count=' +
                str(self._read_throttle_count) + ', write_throttle_count=' +
                str(self._write_throttle_count) + ', storage_throttle_count=' +
                str(self._storage_throttle_count) + ']')

    def get_start_time(self):
        """
        Returns the start time for this usage record in milliseconds since
        the Epoch.

        :returns: the start time.
        :rtype: int
        """
        return self._start_time_ms

    def get_start_time_string(self):
        """
        Returns the start time as an ISO 8601 formatted string. If the start
        timestamp is not set, None is returned.

        :returns: the start time, or None if not set.
        :rtype: str or None
        """
        if self._start_time_ms == 0:
            return None
        return datetime.fromtimestamp(
            float(self._start_time_ms) / 1000).replace(
            tzinfo=tz.UTC).isoformat()

    def get_seconds_in_period(self):
        """
        Returns the number of seconds in this usage record.

        :returns: the number of seconds.
        :rtype: int
        """
        return self._seconds_in_period

    def get_read_units(self):
        """
        Returns the number of read units consumed during this period.

        :returns: the read units.
        :rtype: int
        """
        return self._read_units

    def get_write_units(self):
        """
        Returns the number of write units consumed during this period.

        :returns: the write units.
        :rtype: int
        """
        return self._write_units

    def get_storage_gb(self):
        """
        Returns the amount of storage consumed by the table. This
        information may be out of date as it is not maintained in real time.

        :returns: the size in gigabytes.
        :rtype: int
        """
        return self._storage_gb

    def get_read_throttle_count(self):
        """
        Returns the number of read throttling exceptions on this table in
        the time period.

        :returns: the number of throttling exceptions.
        :rtype: int
        """
        return self._read_throttle_count

    def get_write_throttle_count(self):
        """
        Returns the number of write throttling exceptions on this table in
        the time period.

        :returns: the number of throttling exceptions.
        :rtype: int
        """
        return self._write_throttle_count

    def get_storage_throttle_count(self):
        """
        Returns the number of storage throttling exceptions on this table in
        the time period.

        :returns: the number of throttling exceptions.
        :rtype: int
        """
        return self._storage_throttle_count


class TimeUnit(object):
    """
    The time unit to use.
    """
    HOURS = 1
    """Set TimeUnit.HOURS to use hour as time unit"""
    DAYS = 2
    """Set TimeUnit.DAYS to use day as time unit"""


class TimeToLive(object):
    """
    TimeToLive is a utility class that represents a period of time, similar to
    java.time.Duration in Java, but specialized to the needs of this driver.

    This class is restricted to durations of days and hours. It is only used as
    input related to time to live (TTL) for row instances.

    Construction allows only day and hour durations for efficiency reasons.
    Durations of days are recommended as they result in the least amount of
    storage overhead. Only positive durations are allowed on input.

    :param value: value of time.
    :type value: int
    :param timeunit: unit of time, cannot be None.
    :type timeunit: TimeUnit
    :raises IllegalArgumentException: raises the exception if parameters are not
        expected type.
    """

    def __init__(self, value, timeunit):
        """
        All construction is done via this constructor, which validates the
        arguments.
        """
        CheckValue.check_int(value, 'value')
        if (timeunit is None or timeunit != TimeUnit.DAYS and
                timeunit != TimeUnit.HOURS):
            raise IllegalArgumentException(
                'Invalid time unit in TimeToLive construction. Must be ' +
                'not-none and should be DAYS or HOURS.')
        self._value = value
        self._timeunit = timeunit

    def __str__(self):
        timeunit = 'HOURS' if self._timeunit == TimeUnit.HOURS else 'DAYS'
        return str(self._value) + ' ' + timeunit

    @staticmethod
    def of_hours(hours):
        """
        Creates a duration using a period of hours.

        :param hours: the number of hours in the duration, must be a
            non-negative number.
        :type hours: int
        :returns: the duration.
        :rtype: TimeToLive
        :raises IllegalArgumentException: raises the exception if a negative
            value is provided.
        """
        CheckValue.check_int_ge_zero(hours, 'hours')
        return TimeToLive(hours, TimeUnit.HOURS)

    @staticmethod
    def of_days(days):
        """
        Creates a duration using a period of 24 hour days.

        :param days: the number of days in the duration, must be a non-negative
            number.
        :type days: int
        :returns: the duration.
        :rtype: TimeToLive
        :raises IllegalArgumentException: raises the exception if a negative
            value is provided.
        """
        CheckValue.check_int_ge_zero(days, 'days')
        return TimeToLive(days, TimeUnit.DAYS)

    def to_days(self):
        """
        Returns the number of days in this duration, which may be negative.

        :returns: the number of days.
        :rtype: int
        """
        return (self._value if self._timeunit == TimeUnit.DAYS else
                self._value // 24)

    def to_hours(self):
        """
        Returns the number of hours in this duration, which may be negative.

        :returns: the number of hours.
        :rtype: int
        """
        return (self._value if self._timeunit == TimeUnit.HOURS else
                self._value * 24)

    def to_expiration_time(self, reference_time):
        """
        Returns an absolute time representing the duration plus the absolute
        time reference parameter. If an expiration time from the current time is
        desired the parameter should be the current system time in millisecond.
        If the duration of this object is 0, indicating no expiration time, this
        method will return 0, regardless of the reference time.

        :param reference_time: an absolute time in milliseconds since January
            1, 1970.
        :type reference_time: int
        :returns: time in milliseconds, 0 if this object's duration is 0.
        :rtype: int
        :raises IllegalArgumentException: raises the exception if reference_time
            is not positive.
        """
        CheckValue.check_int_gt_zero(reference_time, 'reference_time')
        if self._value == 0:
            return 0
        hours = 24 if self._timeunit == TimeUnit.DAYS else 1
        return reference_time + hours * self._value * 60 * 60 * 1000

    def get_value(self):
        """
        Returns the numeric duration value.

        :returns: the duration value, independent of unit.
        :rtype: int
        """
        return self._value

    def get_unit(self):
        """
        Returns the time unit used for the duration.

        :returns: the timeunit.
        :rtype: TimeUnit
        """
        return self._timeunit

    def unit_is_days(self):
        return self._timeunit == TimeUnit.DAYS

    def unit_is_hours(self):
        return self._timeunit == TimeUnit.HOURS


class UserInfo(object):
    """
    On-premise only.

    A class that encapsulates the information associated with a user including
    the user id and name in the system.
    """

    def __init__(self, user_id, user_name):
        """
        Constructs an instance of UserInfo as result returned by
        :py:meth:`NoSQLHandle.list_users`.
        """
        self._user_id = user_id
        self._user_name = user_name

    def __str__(self):
        return 'id: ' + self._user_id + ', name: ' + self._user_name

    def get_id(self):
        """
        Returns the id associated with the user.

        :returns: the user id string.
        :rtype: str
        """
        return self._user_id

    def get_name(self):
        """
        Returns the name associated with the user.

        :returns: the user name string.
        :rtype: str
        """
        return self._user_name


class Version(object):
    """
    Version is an opaque class that represents the version of a row in the
    database. It is returned by successful :py:class:`GetRequest` and
    can be used in :py:meth:`PutRequest.set_match_version` and
    :py:meth:`DeleteRequest.set_match_version` to conditionally perform those
    operations to ensure an atomic read-modify-write cycle. This is an opaque
    object from an application perspective.

    Use of Version in this way adds cost to operations so it should be done only
    if necessary.

    :param version: a bytearray.
    :type version: bytearray
    :raises IllegalArgumentException: raises the exception if version is not
        a bytearray.
    """

    def __init__(self, version):
        Version._check_version(version)
        self._version = version

    def get_bytes(self):
        """
        Returns the bytearray from the Version.

        :returns: the bytearray from the Version.
        :rtype: bytearray
        """
        return self._version

    @staticmethod
    def create_version(version):
        """
        Returns an instance of :py:class:`Version`.

        :param version: a bytearray or None.
        :type version: bytearray
        :returns: an instance of Version.
        :rtype: Version
        :raises IllegalArgumentException: raises the exception if version is not
            a bytearray or None.
        """
        if version is None:
            return None
        return Version(version)

    @staticmethod
    def _check_version(version):
        if not isinstance(version, bytearray):
            raise IllegalArgumentException(
                'version must be an bytearray. Got:' + str(version))



#
# Copyright (c) 2018, 2022 Oracle and/or its affiliates. All rights reserved.
#
# Licensed under the Universal Permissive License v 1.0 as shown at
#  https://oss.oracle.com/licenses/upl/
#

from json import loads
from logging import FileHandler, Formatter, WARNING, getLogger
from os import mkdir, path
from ssl import SSLContext, SSLError, create_default_context
from sys import argv

from .client import Client
from .common import CheckValue, UserInfo
from .config import NoSQLHandleConfig
from .exception import IllegalArgumentException, IllegalStateException
from .iam import SignatureProvider
from .kv import StoreAccessTokenProvider
from .operations import (
    DeleteRequest, GetIndexesRequest, GetRequest, GetTableRequest,
    ListTablesRequest, MultiDeleteRequest, PrepareRequest, PutRequest,
    QueryRequest, SystemRequest, SystemStatusRequest, TableRequest,
    TableUsageRequest, WriteMultipleRequest)


class NoSQLHandle(object):
    """
    NoSQLHandle is a handle that can be used to access Oracle NoSQL tables. To
    create a connection represented by NoSQLHandle, request an instance using
    :py:class:`NoSQLHandleConfig`, which allows an application to specify
    default values and other configuration information to be used by the handle.

    The same interface is available to both users of the Oracle NoSQL Database
    Cloud Service and the on-premise Oracle NoSQL Database; however, some
    methods and/or parameters are specific to each environment. The
    documentation has notes about whether a class, method, or parameter is
    environment-specific. Unless otherwise noted they are applicable to both
    environments.

    A handle has memory and network resources associated with it. Consequently,
    the :py:meth:`close` method must be invoked to free up the resources when
    the application is done using the handle. To minimize network activity as
    well as resource allocation and deallocation overheads, it's best to avoid
    repeated creation and closing of handles. For example, creating and closing
    a handle around each operation, would incur large resource allocation
    overheads resulting in poor application performance.

    A handle permits concurrent operations, so a single handle is sufficient to
    access tables in a multi-threaded application. The creation of multiple
    handles incurs additional resource overheads without providing any
    performance benefit.

    With the exception of :py:meth:`close` the operations on this interface
    follow a similar pattern. They accept a Request object containing
    parameters, both required and optional. They return a Result object
    containing results. Operation failures throw exceptions. Unique subclasses
    of Request and Result exist for most operations, containing information
    specific to the operation. All of these operations result in remote calls
    across a network.

    All Request instances support specification of parameters for the operation
    as well as the ability to override default parameters which may have been
    specified in :py:class:`NoSQLHandleConfig`, such as request timeouts, etc.

    Objects returned by methods of this interface can only be used safely by
    one thread at a time unless synchronized externally. Request objects are not
    copied and must not be modified by the application while a method on this
    interface is using them.

    For Error and Exception Handling, on success all methods in this interface
    return Result objects. Errors are thrown as exceptions. Exceptions that may
    be retried may succeed on retry. These are instances of
    :py:class:`RetryableException`. Exceptions that may not be retried and if
    retried, will fail again. Exceptions that may be retried return True for
    :py:meth:`RetryableException.ok_to_retry` while those that may not will
    return False. Examples of retryable exceptions are those which indicate
    resource consumption violations such as
    :py:class:`OperationThrottlingException`. Examples of exceptions that should
    not be retried are :py:class:`IllegalArgumentException`,
    :py:class:`TableNotFoundException`, and any other exception indicating a
    syntactic or semantic error.

    Instances of NoSQLHandle are thread-safe and expected to be shared among
    threads.

    :param config: an instance of NoSQLHandleConfig.
    :type config: NoSQLHandleConfig
    :raises IllegalArgumentException: raises the exception if config is not an
        instance of NoSQLHandleConfig.
    """

    def __init__(self, config):
        if not isinstance(config, NoSQLHandleConfig):
            raise IllegalArgumentException(
                'config must be an instance of NoSQLHandleConfig.')
        logger = self._get_logger(config)
        # config SSLContext first, on-prem authorization provider will reuse the
        # context in NoSQLHandleConfig
        self._config_ssl_context(config)
        self._config_auth_provider(config, logger)
        self._client = Client(config, logger)

    def delete(self, request):
        """
        Deletes a row from a table. The row is identified using a primary key
        value supplied in :py:meth:`DeleteRequest.set_key`.

        By default a delete operation is unconditional and will succeed if the
        specified row exists. Delete operations can be made conditional based
        on whether the :py:class:`Version` of an existing row matches that
        supplied by :py:meth:`DeleteRequest.set_match_version`.

        It is also possible, on failure, to return information about the
        existing row. The row, including it's :py:class:`Version` can be
        optionally returned if a delete operation fails because of a Version
        mismatch. The existing row information will only be returned if
        :py:meth:`DeleteRequest.set_return_row` is True and the operation fails
        because :py:meth:`DeleteRequest.set_match_version` is used and the
        operation fails because the row exists and its version does not match.
        Use of :py:meth:`DeleteRequest.set_return_row` may result in additional
        consumed read capacity. If the operation is successful there will be no
        information returned about the previous row.

        :param request: the input parameters for the operation.
        :type request: DeleteRequest
        :returns: the result of the operation.
        :rtype: DeleteResult
        :raises IllegalArgumentException: raises the exception if request is not
            an instance of :py:class:`DeleteRequest`.
        :raises NoSQLException: raises the exception if the operation cannot be
            performed for any other reason.
        """
        if not isinstance(request, DeleteRequest):
            raise IllegalArgumentException(
                'The parameter should be an instance of DeleteRequest.')
        return self._execute(request)

    def get(self, request):
        """
        Gets the row associated with a primary key. On success the value of the
        row is available using the :py:meth:`GetResult.get_value` operation. If
        there are no matching rows that method will return None.

        The default consistency used for the operation is Consistency.EVENTUAL
        unless an explicit value has been set using
        :py:meth:`NoSQLHandleConfig.set_consistency` or
        :py:meth:`GetRequest.set_consistency`. Use of Consistency.ABSOLUTE may
        affect latency of the operation and may result in additional cost for
        the operation.

        :param request: the input parameters for the operation.
        :type request: GetRequest
        :returns: the result of the operation.
        :rtype: GetResult
        :raises IllegalArgumentException: raises the exception if request is not
            an instance of :py:class:`GetRequest`.
        :raises NoSQLException: raises the exception if the operation cannot be
            performed for any other reason.
        """
        if not isinstance(request, GetRequest):
            raise IllegalArgumentException(
                'The parameter should be an instance of GetRequest.')
        return self._execute(request)

    def get_indexes(self, request):
        """
        Returns information about and index, or indexes on a table. If no index
        name is specified in the :py:class:`GetIndexesRequest`, then information
        on all indexes is returned.

        :param request: the input parameters for the operation.
        :type request: GetIndexesRequest
        :returns: the result of the operation.
        :rtype: GetIndexesResult
        :raises IllegalArgumentException: raises the exception if request is not
            an instance of :py:class:`GetIndexesRequest`.
        :raises NoSQLException: raises the exception if the operation cannot be
            performed for any other reason.
        """
        if not isinstance(request, GetIndexesRequest):
            raise IllegalArgumentException(
                'The parameter should be an instance of GetIndexesRequest.')
        return self._execute(request)

    def get_table(self, request):
        """
        Gets static information about the specified table including its state,
        provisioned throughput and capacity and schema. Dynamic information such
        as usage is obtained using :py:meth:`get_table_usage`. Throughput,
        capacity and usage information is only available when using the Cloud
        Service and will be None or not defined on-premise.

        :param request: the input parameters for the operation.
        :type request: GetTableRequest
        :returns: the result of the operation.
        :rtype: TableResult
        :raises IllegalArgumentException: raises the exception if request is not
            an instance of :py:class:`GetTableRequest`.
        :raises TableNotFoundException: raises the exception if the specified
            table does not exist.
        :raises NoSQLException: raises the exception if the operation cannot be
            performed for any other reason.
        """
        if not isinstance(request, GetTableRequest):
            raise IllegalArgumentException(
                'The parameter should be an instance of GetTableRequest.')
        res = self._execute(request)
        # Update rate limiters, if table has limits.
        self._client.update_rate_limiters(
            res.get_table_name(), res.get_table_limits())
        return res

    def get_table_usage(self, request):
        """
        Cloud service only.

        Gets dynamic information about the specified table such as the current
        throughput usage. Usage information is collected in time slices and
        returned in individual usage records. It is possible to specify a
        time-based range of usage records using input parameters.

        :param request: the input parameters for the operation.
        :type request: TableUsageRequest
        :returns: the result of the operation.
        :rtype: TableUsageResult
        :raises IllegalArgumentException: raises the exception if request is not
            an instance of :py:class:`TableUsageRequest`.
        :raises TableNotFoundException: raises the exception if the specified
            table does not exist.
        :raises NoSQLException: raises the exception if the operation cannot be
            performed for any other reason.
        """
        if not isinstance(request, TableUsageRequest):
            raise IllegalArgumentException(
                'The parameter should be an instance of TableUsageRequest.')
        return self._execute(request)

    def list_tables(self, request):
        """
        Lists tables, returning table names. If further information about a
        specific table is desired the :py:meth:`get_table` interface may be
        used. If a given identity has access to a large number of tables the
        list may be paged using input parameters.

        :param request: the input parameters for the operation.
        :type request: ListTablesRequest
        :returns: the result of the operation.
        :rtype: ListTablesResult
        :raises IllegalArgumentException: raises the exception if request is not
            an instance of :py:class:`ListTablesRequest`.
        :raises NoSQLException: raises the exception if the operation cannot be
            performed for any other reason.
        """
        if not isinstance(request, ListTablesRequest):
            raise IllegalArgumentException(
                'The parameter should be an instance of ListTablesRequest.')
        return self._execute(request)

    def multi_delete(self, request):
        """
        Deletes multiple rows from a table in an atomic operation. The key used
        may be partial but must contain all of the fields that are in the shard
        key. A range may be specified to delete a range of keys.

        :param request: the input parameters for the operation.
        :type request: MultiDeleteRequest
        :returns: the result of the operation.
        :rtype: MultiDeleteResult
        :raises IllegalArgumentException: raises the exception if request is not
            an instance of :py:class:`MultiDeleteRequest`.
        :raises NoSQLException: raises the exception if the operation cannot be
            performed for any other reason.
        """
        if not isinstance(request, MultiDeleteRequest):
            raise IllegalArgumentException(
                'The parameter should be an instance of MultiDeleteRequest.')
        return self._execute(request)

    def prepare(self, request):
        """
        Prepares a query for execution and reuse. See :py:meth:`query` for
        general information and restrictions. It is recommended that prepared
        queries are used when the same query will run multiple times as
        execution is much more efficient than starting with a query string every
        time. The query language and API support query variables to assist with
        re-use.

        :param request: the input parameters for the operation.
        :type request: PrepareRequest
        :returns: the result of the operation.
        :rtype: PrepareResult
        :raises IllegalArgumentException: raises the exception if request is not
            an instance of :py:class:`PrepareRequest`.
        :raises NoSQLException: raises the exception if the operation cannot be
            performed for any other reason.
        """
        if not isinstance(request, PrepareRequest):
            raise IllegalArgumentException(
                'The parameter should be an instance of PrepareRequest.')
        return self._execute(request)

    def put(self, request):
        """
        Puts a row into a table. This method creates a new row or overwrites
        an existing row entirely. The value used for the put is in the
        :py:class:`PutRequest` object and must contain a complete primary key
        and all required fields.

        It is not possible to put part of a row.
        Any fields that are not provided will be defaulted, overwriting any
        existing value. Fields that are not noneable or defaulted must be
        provided or an exception will be thrown.

        By default a put operation is unconditional, but put operations can be
        conditional based on existence, or not, of a previous value as well as
        conditional on the :py:class:`Version` of the existing value.

            Use PutOption.IF_ABSENT to do a put only if there is no existing row
            that matches the primary key.\n
            Use PutOption.IF_PRESENT to do a put only if there is an existing
            row that matches the primary key.\n
            Use PutOption.IF_VERSION to do a put only if there is an existing
            row that matches the primary key and its :py:class:`Version` matches
            that provided.

        It is also possible, on failure, to return information about the
        existing row. The row, including it's :py:class:`Version` can be
        optionally returned if a put operation fails because of a Version
        mismatch or if the operation fails because the row already exists.
        The existing row information will only be returned if
        :py:meth:`PutRequest.set_return_row` is True and one of the following
        occurs:

            The PutOption.IF_ABSENT is used and the operation fails because the
            row already exists.\n
            The PutOption.IF_VERSION is used and the operation fails because the
            row exists and its version does not match.

        Use of :py:meth:`PutRequest.set_return_row` may result in additional
        consumed read capacity. If the operation is successful there will be no
        information returned about the previous row.

        :param request: the input parameters for the operation.
        :type request: PutRequest
        :returns: the result of the operation.
        :rtype: PutResult
        :raises IllegalArgumentException: raises the exception if request is not
            an instance of :py:class:`PutRequest`.
        :raises NoSQLException: raises the exception if the operation cannot be
            performed for any other reason.
        """
        if not isinstance(request, PutRequest):
            raise IllegalArgumentException(
                'The parameter should be an instance of PutRequest.')
        return self._execute(request)

    def query(self, request):
        """
        Queries a table based on the query statement specified in the
        :py:class:`QueryRequest`.

        Queries that include a full shard key will execute much more efficiently
        than more distributed queries that must go to multiple shards.

        Table and system-style queries such as "CREATE TABLE ..." or "DROP TABLE
        ..." are not supported by this interfaces. Those operations must be
        performed using :py:meth:`table_request` or :py:meth:`system_request` as
        appropriate.

        The amount of data read by a single query request is limited by a system
        default and can be further limited using
        :py:meth:`QueryRequest.set_max_read_kb`. This limits the amount of data
        *read* and not the amount of data *returned*, which means that a query
        can return zero results but still have more data to read. This situation
        is detected by checking if the :py:class:`QueryRequest` is done using
        :py:meth:`QueryRequest.is_done`. For this reason queries should always
        operate in a loop, acquiring more results, until
        :py:meth:`QueryRequest.is_done` returns True, indicating that the query
        is done.

        :param request: the input parameters for the operation.
        :type request: QueryRequest
        :returns: the result of the operation.
        :rtype: QueryResult
        :raises IllegalArgumentException: raises the exception if request is not
            an instance of :py:class:`QueryRequest`.
        :raises NoSQLException: raises the exception if the operation cannot be
            performed for any other reason.
        """
        if not isinstance(request, QueryRequest):
            raise IllegalArgumentException(
                'The parameter should be an instance of QueryRequest.')
        return self._execute(request)

    def system_request(self, request):
        """
        On-premise only.

        Performs a system operation on the system, such as administrative
        operations that don't affect a specific table. For table-specific
        operations use :py:meth:`table_request` or :py:meth:`do_table_request`.

        Examples of statements in the :py:class:`SystemRequest` passed to this
        method include:

            CREATE NAMESPACE mynamespace\n
            CREATE USER some_user IDENTIFIED BY password\n
            CREATE ROLE some_role\n
            GRANT ROLE some_role TO USER some_user

        This operation is implicitly asynchronous. The caller must poll using
        methods on :py:class:`SystemResult` to determine when it has completed.

        :param request: the input parameters for the operation.
        :type request: SystemRequest
        :returns: the result of the operation.
        :rtype: SystemResult
        :raises IllegalArgumentException: raises the exception if request is not
            an instance of :py:class:`SystemRequest`.
        :raises NoSQLException: raises the exception if the operation cannot be
            performed for any other reason.
        """
        if not isinstance(request, SystemRequest):
            raise IllegalArgumentException(
                'The parameter should be an instance of SystemRequest.')
        return self._execute(request)

    def system_status(self, request):
        """
        On-premise only.

        Checks the status of an operation previously performed using
        :py:meth:`system_request`.

        :param request: the input parameters for the operation.
        :type request: SystemStatusRequest
        :returns: the result of the operation.
        :rtype: SystemResult
        :raises IllegalArgumentException: raises the exception if request is not
            an instance of :py:class:`SystemStatusRequest`.
        :raises NoSQLException: raises the exception if the operation cannot be
            performed for any other reason.
        """
        if not isinstance(request, SystemStatusRequest):
            raise IllegalArgumentException(
                'The parameter should be an instance of SystemStatusRequest.')
        return self._execute(request)

    def table_request(self, request):
        """
        Performs an operation on a table. This method is used for creating and
        dropping tables and indexes as well as altering tables. Only one
        operation is allowed on a table at any one time.

        This operation is implicitly asynchronous. The caller must poll using
        methods on :py:class:`TableResult` to determine when it has completed.

        :param request: the input parameters for the operation.
        :type request: TableRequest
        :returns: the result of the operation.
        :rtype: TableResult
        :raises IllegalArgumentException: raises the exception if request is not
            an instance of :py:class:`TableRequest`.
        :raises NoSQLException: raises the exception if the operation cannot be
            performed for any other reason.
        """
        if not isinstance(request, TableRequest):
            raise IllegalArgumentException(
                'The parameter should be an instance of TableRequest.')
        res = self._execute(request)
        # Update rate limiters, if table has limits.
        self._client.update_rate_limiters(
            res.get_table_name(), res.get_table_limits())
        return res

    def write_multiple(self, request):
        """
        Executes a sequence of operations associated with a table that share the
        same shard key portion of their primary keys, all the specified
        operations are executed within the scope of a single transaction.

        There are some size-based limitations on this operation:

            The max number of individual operations (put, delete) in a single
            WriteMultipleRequest is 50.\n
            The total request size is limited to 25MB.

        :param request: the input parameters for the operation.
        :type request: WriteMultipleRequest
        :returns: the result of the operation.
        :rtype: WriteMultipleResult
        :raises IllegalArgumentException: raises the exception if request is not
            an instance of :py:class:`WriteMultipleRequest`.
        :raises RowSizeLimitException: raises the exception if data size in an
            operation exceeds the limit.
        :raises BatchOperationNumberLimitException: raises the exception if the
            number of operations exceeds this limit.
        :raises NoSQLException: raises the exception if the operation cannot be
            performed for any other reason.
        """
        if not isinstance(request, WriteMultipleRequest):
            raise IllegalArgumentException(
                'The parameter should be an instance of WriteMultipleRequest.')
        return self._execute(request)

    def close(self):
        """
        Close the NoSQLHandle.
        """
        if self._client is not None:
            self._client.shut_down()
            self._client = None

    def do_system_request(self, statement, timeout_ms=30000,
                          poll_interval_ms=1000):
        """
        On-premise only.

        A convenience method that performs a SystemRequest and waits for
        completion of the operation. This is the same as calling
        :py:meth:`system_request` then calling
        :py:meth:`SystemResult.wait_for_completion`. If the operation fails an
        exception is thrown.

        System requests are those related to namespaces and security and are
        generally independent of specific tables. Examples of statements include

            CREATE NAMESPACE mynamespace\n
            CREATE USER some_user IDENTIFIED BY password\n
            CREATE ROLE some_role\n
            GRANT ROLE some_role TO USER some_user

        :param statement: the system statement for the operation.
        :type statement: str
        :param timeout_ms: the amount of time to wait for completion, in
            milliseconds.
        :type timeout_ms: int
        :param poll_interval_ms: the polling interval for the wait operation.
        :type poll_interval_ms: int
        :returns: the result of the system request.
        :rtype: SystemResult
        :raises IllegalArgumentException: raises the exception if any of the
            parameters are invalid or required parameters are missing.
        :raises RequestTimeoutException: raises the exception if the operation
            times out.
        :raises NoSQLException: raises the exception if the operation cannot be
            performed for any other reason.
        """
        CheckValue.check_str(statement, 'statement')
        req = SystemRequest().set_statement(statement)
        res = self.system_request(req)
        res.wait_for_completion(self, timeout_ms, poll_interval_ms)
        return res

    def do_table_request(self, request, timeout_ms, poll_interval_ms):
        """
        A convenience method that performs a TableRequest and waits for
        completion of the operation. This is the same as calling
        :py:meth:`table_request` then calling
        :py:meth:`TableResult.wait_for_completion`. If the operation fails an
        exception is thrown. All parameters are required.

        :param request: the :py:class:`TableRequest` to perform.
        :type request: TableRequest
        :param timeout_ms: the amount of time to wait for completion, in
            milliseconds.
        :type timeout_ms: int
        :param poll_interval_ms: the polling interval for the wait operation.
        :type poll_interval_ms: int
        :returns: the result of the table request.
        :rtype: TableResult
        :raises IllegalArgumentException: raises the exception if any of the
            parameters are invalid or required parameters are missing.
        :raises RequestTimeoutException: raises the exception if the operation
            times out.
        :raises NoSQLException: raises the exception if the operation cannot be
            performed for any other reason.
        """
        if not isinstance(request, TableRequest):
            raise IllegalArgumentException(
                'The parameter should be an instance of TableRequest.')
        res = self.table_request(request)
        res.wait_for_completion(self, timeout_ms, poll_interval_ms)
        return res

    def list_namespaces(self):
        """
        On-premise only.

        Returns the namespaces in a store as a list of string.

        :returns: the namespaces, or None if none are found.
        :rtype: list(str)
        """
        res = self.do_system_request('show as json namespaces')
        json_res = res.get_result_string()
        if json_res is None:
            return None
        root = loads(json_res)
        namespaces = root.get('namespaces')
        if namespaces is None:
            return None
        results = list()
        for namespace in namespaces:
            results.append(namespace)
        return results

    def list_roles(self):
        """
        On-premise only.

        Returns the roles in a store as a list of string.

        :returns: the list of roles, or None if none are found.
        :rtype: list(str)
        """
        res = self.do_system_request('show as json roles')
        json_res = res.get_result_string()
        if json_res is None:
            return None
        root = loads(json_res)
        roles = root.get('roles')
        if roles is None:
            return None
        results = list()
        for role in roles:
            results.append(role['name'])
        return results

    def list_users(self):
        """
        On-premise only.

        Returns the users in a store as a list of :py:class:`UserInfo`.

        :returns: the list of users, or None if none are found.
        :rtype: list(UserInfo)
        """
        res = self.do_system_request('show as json users')
        json_res = res.get_result_string()
        if json_res is None:
            return None
        root = loads(json_res)
        users = root.get('users')
        if users is None:
            return None
        results = list()
        for user in users:
            results.append(UserInfo(user['id'], user['name']))
        return results

    def get_client(self):
        # For testing use
        return self._client

    def _execute(self, request):
        # Ensure that the client exists and hasn't been closed.
        if self._client is None:
            raise IllegalStateException('NoSQLHandle has been closed.')
        return self._client.execute(request)

    def _get_logger(self, config):
        """
        Returns the logger used for the driver. If no logger is specified,
        create one based on this class name.
        """
        if config.get_logger() is None and config.is_default_logger():
            logger = getLogger(self.__class__.__name__)
            logger.setLevel(WARNING)
            log_dir = path.join(path.abspath(path.dirname(argv[0])), 'logs')
            if not path.exists(log_dir):
                mkdir(log_dir)
            handler = FileHandler(path.join(log_dir, 'driver.log'))
            formatter = Formatter('%(asctime)s [%(levelname)s] %(message)s')
            handler.setFormatter(formatter)
            logger.addHandler(handler)
        else:
            logger = config.get_logger()
        return logger

    @staticmethod
    def _config_auth_provider(config, logger):
        provider = config.get_authorization_provider()
        if provider.get_logger() is None:
            provider.set_logger(logger)
        if (isinstance(provider, StoreAccessTokenProvider) and
                provider.is_secure()):
            if provider.get_endpoint() is None:
                endpoint = config.get_service_url().geturl()
                if endpoint.endswith('/'):
                    endpoint = endpoint[:len(endpoint) - 1]
                provider.set_endpoint(endpoint)
            provider.set_ssl_context(config.get_ssl_context())
        elif isinstance(provider, SignatureProvider):
            provider.set_service_url(config)

    @staticmethod
    def _config_ssl_context(config):
        if config.get_ssl_context() is not None:
            return
        if config.get_service_url().scheme == 'https':
            try:
                if config.get_ssl_protocol() is None:
                    ctx = create_default_context()
                else:
                    ctx = SSLContext(config.get_ssl_protocol())
                if config.get_ssl_cipher_suites() is not None:
                    ctx.set_ciphers(config.get_ssl_cipher_suites())
                if config.get_ssl_ca_certs() is not None:
                    ctx.load_verify_locations(config.get_ssl_ca_certs())
                config.set_ssl_context(ctx)
            except (SSLError, ValueError) as err:
                raise IllegalArgumentException(str(err))

    def get_stats_control(self):
        return self._client.get_stats_control()



#
# Copyright (c) 2018, 2022 Oracle and/or its affiliates. All rights reserved.
#
# Licensed under the Universal Permissive License v 1.0 as shown at
#  https://oss.oracle.com/licenses/upl/
#

import pprint
import sys
import uuid
from collections import Callable
from datetime import datetime
from logging import INFO
from threading import Timer, Lock
from time import localtime
from typing import Any, Dict

from . import StatsProfile
from .exception import SecurityInfoNotReadyException, ThrottlingException, \
    IllegalArgumentException
from .common import LogUtils, synchronized, CheckValue
from .kv.exception import AuthenticationException
from .version import __version__


class StatsControl:
    """
    StatsControl allows user to control the collection of driver statistics at
    runtime.

    The statistics data is collected for an interval of time. At the end of the
    interval, the stats data is logged in a specified JSON format that can be
    filtered and parsed. After the logging, the counters are cleared and
    collection of data resumes.

    Collection intervals are aligned to the top of the hour. This means first
    interval logs may contain stats for a shorter interval.

    Collection of stats are controlled by the following environment variables:

      NOSQL_STATS_PROFILE=[none|regular|more|all]

         Specifies the stats profile:

          * none - disabled,
          * regular - per request: counters, errors, latencies, delays, retries.
            This incurs minimum overhead.
          * more - stats above with 95th and 99th percentile latencies.
            This may add 0.5% overhead compared to none stats profile.
          * all - stats above with per query information.
            This may add 1% overhead compared to none stats profile.

      NOSQL_STATS_INTERVAL=600 Interval in seconds to log the stats, by
      default is 10 minutes.

      NOSQL_STATS_PRETTY_PRINT=true Option to enable pretty printing of the JSON
      data, default value is false.

    Collection of stats can also be used by using the API:

    :py:meth:`NoSQLHandleConfig.set_stats_profile` or
    :py:meth:`StatsControl.set_profile`. At runtime stats collection can be
    enabled selectively by using :py:meth:`StatsControl.start` ond
    :py:meth:`StatsControl.stop`. The following example shows how to use a stats
    handler and how to control the stas at runtime::

        def stats_handler(stats):
            # type: (Dict) -> None
            print("Stats : " + str(stats))
        ...
        config = NoSQLHandleConfig( endpoint )
        config.set_stats_profile(StatsProfile.REGULAR)
        config.set_stats_interval(600)
        config.set_stats_pretty_print(False)
        config.set_stats_handler(stats_handler)

        handle = NoSQLHandle(config)

        handle = get_handle(tenant_id)

        stats_control = handle.get_stats_control()

        #... application code without stats

        # enable observations
        stats_control.start();

        #... application code with REGULAR stats

        # For particular parts of code profile can be changed to collect more stats.
        stats_control.set_stats_profile(StatsProfile.ALL)
        #... more sensitive code with ALL stats

        stats_control.set_stats_profile(StatsProfile.REGULAR)
        #... application code with REGULAR stats

        # disable observations
        stats_control.stop()

        #... application code without stats
        handle.close()

    The following is an example of stats log entry using the ALL
    profile:

    - A one time entry containing stats id and options::

        INFO: Client stats|{    // INFO log entry
        "sdkName" : "Oracle NoSQL SDK for Python",  // SDK name
        "sdkVersion" : "5.2.4",                 // SDK version
        "clientId" : "f595b333",                  // NoSQLHandle id
        "profile" : "ALL",                        // stats profile
        "intervalSec" : 600,                      // interval length in seconds
        "prettyPrint" : true,                     // JSON pretty print
        "rateLimitingEnabled" : false}            // if rate limiting is enabled

    - An entry at the end of each interval containing the stats values::

         ``INFO: Client stats|{
         "clientId" : "b7bc7734",              // id of NoSQLHandle object
         "startTime" : "2021-09-20T20:11:42Z", // UTC start interval time
         "endTime" : "2021-09-20T20:11:47Z",   // UTC end interval time
         "requests" : [{                       // array of types of requests
           "name" : "Get",                       // stats for GET request type
           "httpRequestCount" : 2,               // count of http requests
           "errors" : 0,                         // number of errors in interval
           "httpRequestLatencyMs" : {            // response time of http requests
             "min" : 4,                            // minimum value in interval
             "avg" : 4.5,                          // average value in interval
             "max" : 5,                            // maximum value in interval
             "95th" : 5,                           // 95th percentile value
             "99th" : 5                            // 99th percentile value
           },
           "requestSize" : {                     // http request size in bytes
             "min" : 42,                           // minimum value in interval
             "avg" : 42.5,                         // average value in interval
             "max" : 43                            // maximum value in interval
           },
           "resultSize" : {                      // http result size in bytes
             "min" : 193,                          // minimum value in interval
             "avg" : 206.5,                        // average value in interval
             "max" : 220                           // maximum value in interval
           },
           "rateLimitDelayMs" : 0,               // delay in milliseconds introduced by the rate limiter
           "retry" : {                           // retries
             "delayMs" : 0,                        // delay in milliseconds introduced by retries
             "authCount" : 0,                      // no of auth retries
             "throttleCount" : 0,                  // no of throttle retries
             "count" : 0                           // total number of retries
           }
         }, {
           "name" : "Query",                   // stats for all QUERY type requests
           "httpRequestCount" : 14,
           "errors" : 0,
           "httpRequestLatencyMs" : {
             "min" : 3,
             "avg" : 13.0,
             "max" : 32,
             "95th" : 32,
             "99th" : 32
           },
           "resultSize" : {
             "min" : 146,
             "avg" : 7379.71,
             "max" : 10989
           },
           "requestSize" : {
             "min" : 65,
             "avg" : 709.85,
             "max" : 799
           },
           "rateLimitDelayMs" : 0,
           "retry" : {
             "delayMs" : 0,
             "authCount" : 0,
             "throttleCount" : 0,
             "count" : 0
           }
         }, {
           "name" : "Put",                    // stats for PUT type requests
           "httpRequestCount" : 1002,
           "errors" : 0,
           "httpRequestLatencyMs" : {
             "min" : 1,
             "avg" : 4.41,
             "max" : 80,
             "95th" : 8,
             "99th" : 20
           },
           "requestSize" : {
             "min" : 90,
             "avg" : 90.16,
             "max" : 187
           },
           "resultSize" : {
             "min" : 58,
             "avg" : 58.0,
             "max" : 58
           },
           "rateLimitDelayMs" : 0,
           "retry" : {
             "delayMs" : 0,
             "authCount" : 0,
             "throttleCount" : 0,
             "count" : 0
           }
         }],
         "queries" : [{            // query stats aggregated by query statement
                                     // query statement
           "query" : "SELECT * FROM audienceData ORDER BY cookie_id",
                                     // query plan description
           "plan" : "SFW([6])
              [
                FROM:
                  RECV([3])
                    [
                      DistributionKind : ALL_PARTITIONS,
                      Sort Fields : sort_gen,
                    ] as $from-0
                SELECT:
                 FIELD_STEP([6])
                   [
                    VAR_REF($from-0)([3]),
                    audienceData
                   ]
              ]",
           "doesWrites" : false,
           "httpRequestCount" : 12,  // number of http calls to the server
           "unprepared" : 1,         // number of query requests without prepare
           "simple" : false,         // type of query
           "count" : 20,             // number of handle.query() API calls
           "errors" : 0,             // number of calls trowing exception
           "httpRequestLatencyMs" : {// response time of http requests in milliseconds
             "min" : 8,                // minimum value in interval
             "avg" : 14.58,            // average value in interval
             "max" : 32,               // maximum value in interval
             "95th" : 32,              // 95th percentile value in interval
             "99th" : 32               // 99th percentile value in interval
           },
           "requestSize" : {         // http request size in bytes
             "min" : 65,               // minimum value in interval
             "avg" : 732.5,            // average value in interval
             "max" : 799               // maximum value in interval
           },
           "resultSize" : {          // http result size in bytes
             "min" : 914,              // minimum value in interval
             "avg" : 8585.33,          // average value in interval
             "max" : 10989             // maximum value in interval
           },
           "rateLimitDelayMs" : 0,   // total delay introduced by rate limiter in milliseconds
           "retry" : {               // automatic retries
             "delayMs" : 0,            // delay introduced by retries
             "authCount" : 0,          // count of auth related retries
             "throttleCount" : 0,      // count of throttle related retries
             "count" : 0               // total count of retries
           }
         }]
        }``

   The log entries go to the logger configured in NoSQLHandlerConfig. By
   default, if no logger is configured the statistics entries, if enabled,
   will be logged to file **logs/driver.log** in the local directory.

   Stats collection is not dependent of logging configuration, even if
   logging is disabled, collection of stats will still happen if stats
   profile other than *none* is used. In this case, the stats are available by
   using the stats handler.

   Depending on the type of query, if client processing is required, for
   example in the case of ordered or aggregate queries, indicated by the
   false **simple** field of the **query** entry, the **count** and
   **httpRequestsCount** numbers will differ. **count** represents
   the number of ``handle.query()`` API calls and **httpRequestCount**
   represents the number of internal http requests from server. For these
   type of queries, the driver executes several simpler queries, per
   shard or partition, and than combines the results locally.

    Note: connection statistics are not available for NoSQL Python driver.
    """
    LOG_PREFIX = "Client stats|"

    def __init__(self, config, logger, is_rate_limiting_enabled):
        self._logutils = LogUtils(logger)
        self._is_rate_limiting_enabled = is_rate_limiting_enabled

        self._profile = config.get_stats_profile()
        self._interval = config.get_stats_interval()
        self._pretty_print = config.get_stats_pretty_print()
        self._stats_handler = config.get_stats_handler()

        self._enable_collection = False

        self._stats = None   # type: Stats
        self._id = str(uuid.uuid4())[:8]

        if self._profile is not StatsProfile.NONE:
            self._logutils.set_level(INFO)
            self._logutils.log_info(StatsControl.LOG_PREFIX +
                                    "{\"sdkName\": \"Oracle NoSQL SDK for "
                                    "Python" +
                                    "\", \"sdkVersion\": \"" + __version__ +
                                    "\", \"clientId\": \"" + self._id +
                                    "\", \"profile\": \"" +
                                    str(self._profile.name) +
                                    "\", \"intervalSec\": " +
                                    str(self._interval) +
                                    ", \"prettyPrint\": " +
                                    str(self._pretty_print) +
                                    ", \"rateLimitingEnabled\": " +
                                    str(self._is_rate_limiting_enabled) +
                                    "}")

        self.start()

    def set_stats_handler(self, stats_handler):
        # type: (Callable[Dict]) -> StatsControl
        """
        Registers a user defined stats handler. The handler is called at the end
        of the interval with a structure containing the logged stat values.
        """
        if not isinstance(stats_handler, Callable):
            raise IllegalArgumentException(
                'stats_handler must be of Callable[Dict] type')
        self._stats_handler = stats_handler
        return self

    def get_stats_handler(self):
        # type: (...) -> Callable[Dict]
        """
        Returns the registered handler.
        """
        return self._stats_handler

    def start(self):
        """
        Collection of stats is enabled only between start and stop or from the
        beginning if environment property NOSQL_STATS_PROFILE is not "none".
        """
        if self._profile is StatsProfile.NONE:
            if self._stats is not None:
                self._stats.shutdown()
            self._stats = None
        else:
            if self._stats is None:
                self._stats = Stats(self)
            self._enable_collection = True

    def stop(self):
        """
        Stops collection of stats.
        """
        self._enable_collection = False

    def is_started(self):
        """
        Returns true if collection of stats is enabled, otherwise returns false.
        """
        return self._enable_collection

    def shutdown(self):
        """
        Logs the stats collected and stops the timer.
        """
        if self._stats is not None:
            self._stats.shutdown()

    def get_profile(self):
        # type: () -> StatsProfile
        """
        Returns the stats collection profile. Default stats profile is NONE.
        """
        return self._profile

    def set_profile(self, profile):
        # type: (StatsProfile) -> StatsControl
        """
        Set the stats collection stats_profile. Default stats stats_profile is
        NONE.
        """
        if profile is not None and not isinstance(profile, StatsProfile):
            raise IllegalArgumentException(
                'stats_profile must be a StatsProfile.')
        self._profile = profile
        return self

    def get_interval(self):
        # type: () -> int
        """
        Returns the current collection interval.
        Default interval is 600 seconds, i.e. 10 min.
        """
        return self._interval

    def get_id(self):
        # type: () -> str
        """
        Returns a pseudo unique string to identify the NoSQLHandle object.
        """
        return self._id

    def get_pretty_print(self):
        """
        Returns the current JSON pretty print flag.
        Default is disabled.
        """
        return self._pretty_print

    def set_pretty_print(self, pretty_print):
        # type: (bool) -> StatsControl
        """
        Enable JSON pretty print for easier human reading.
        Default is disabled.
        """
        CheckValue.check_boolean(pretty_print, "pretty_print")
        self._pretty_print = pretty_print
        return self

    def get_logger(self):
        """
        Returns the current logger.
        """
        if self._logutils is not None:
            return self._logutils.get_logger()
        else:
            return None

    def observe(self, request, req_size, res_size, network_latency):
        """
        Internal method only.
        """
        if self._enable_collection and self._stats is not None:
            self._stats.observe(request, False, req_size, res_size,
                                network_latency)

    def observe_error(self, request):
        """
        Internal method only.
        """
        if self._enable_collection and self._stats is not None:
            self._stats.observe_error(request)

    def observe_query(self, query_request):
        """
        Internal method only.
        """
        if self._enable_collection and self._stats is not None:
            self._stats.observe_query(query_request)


class RepeatedTimer(object):
    """
    Class that implements the timer to log stats at certain interval. The first
    event is called after the delay and the rest are called after each interval.
    """

    def __init__(self, delay, interval, function, *args, **kwargs):
        # type: (float, int, Callable, [Any], [[Any]]) -> None
        self._timer = None
        self._delay = delay
        self._interval = interval
        self._function = function
        self._args = args
        self._kwargs = kwargs
        self._firstTime = True
        self._is_running = False
        self.start()

    def _run(self):
        self._is_running = False
        self.start()
        self._function(*self._args, **self._kwargs)

    def start(self):
        if not self._is_running:
            interval = self._interval
            if self._firstTime:
                interval = self._delay
                self._firstTime = False
            self._timer = Timer(interval, self._run)
            self._timer.start()
            self._is_running = True

    def stop(self):
        self._timer.cancel()
        self._is_running = False


class QueryEntryStat:
    """
    Statistics for a certain query.
    """

    def __init__(self, profile, query_request):
        self._plan = None
        self._does_writes = False
        self._count = 0
        self._unprepared = 0
        self._simple = False
        self._req_stats = ReqStats(profile)

    def observe_query(self, query_request):
        # type: (QueryRequest) -> None
        self._count += 1
        if not query_request.is_prepared():
            self._unprepared += 1
        if query_request.is_prepared() and query_request.is_simple_query():
            self._simple = True

    def observe(self, error, retries, retry_delay,
                rate_limit_delay, auth_count, throttle_count,
                req_size, res_size, network_latency):
        # type: (bool, int, int, int, int, int, int, int, int) -> None
        self._req_stats.observe(error, retries, retry_delay, rate_limit_delay,
                                auth_count, throttle_count, req_size, res_size,
                                network_latency)

    def to_json(self, queries, query):
        # type: ([Any], str) -> None
        q = {"query": query, "count": self._count,
             "unprepared": self._unprepared, "simple": self._simple,
             "doesWrites": self._does_writes}
        if self._plan is not None:
            q["plan"] = self._plan

        self._req_stats.to_map_value(q)
        queries.append(q)

    def get_plan(self):
        return self._plan


class ExtraQueryStats:
    """
    Statistics for all queries.
    """

    def __init__(self, profile):
        self._queries = {}  # type: Dict[str, QueryEntryStat]
        self._profile = profile

    def observe_query(self, query_request):
        q_stat = self.get_extra_query_stat_entry(query_request)
        q_stat.observe_query(query_request)

    def observe_q_rec(self, query_request, error, retries,
                      retry_delay, rate_limit_delay, auth_count,
                      throttle_count, req_size, res_size,
                      network_latency):
        """
        type: (borneo.Request, bool, int, int, int, int, int, int, int, int) ->
          None
        """
        q_stat = self.get_extra_query_stat_entry(query_request)
        q_stat.observe(error, retries, retry_delay, rate_limit_delay,
                       auth_count, throttle_count, req_size, res_size,
                       network_latency)

    def get_extra_query_stat_entry(self, query_request):
        # type: (borneo.QueryRequest) -> QueryEntryStat
        sql = query_request.get_statement()
        if sql is None and self._prepared_statement is not None:
            sql = self._prepared_statement.get_sql_text()
        q_stat = self._queries.get(sql)
        if q_stat is None:
            q_stat = QueryEntryStat(self._profile, query_request)
            self._queries[sql] = q_stat
        if q_stat.get_plan() is None:
            if query_request.get_prepared_statement() is not None:
                prep_stmt = query_request.get_prepared_statement()
                q_stat._plan = prep_stmt.print_driver_plan()
                q_stat._does_writes = prep_stmt.does_writes()
        return q_stat

    def to_json(self, root):
        if len(self._queries) > 0:
            queries = []
            root["queries"] = queries
            for k in self._queries:
                self._queries[k].to_json(queries, k)

    def clear(self):
        self._queries = {}  # type: Dict[str, QueryEntryStat]


class Percentile:
    """
    Implements storing and computation of percentiles for a given set of values.
    """

    def __init__(self):
        self._values = []    # type: [int]

    def add_value(self, network_latency):
        # type: (int) -> None
        self._values.append(network_latency)

    def get_percentile(self, percentiles):
        # type: ([float]) -> [int]
        size = len(self._values)
        self._values.sort()
        # if no values available return -1
        # if requested percentile is 0 or less return first value
        # if 1 or more return last value
        # else return corresponding value
        return [-1 if size == 0 else
                self._values[0] if round(v * size) <= 0 else
                self._values[size - 1] if round(v * size) >= size
                else self._values[round(v * size) - 1]
                for v in percentiles]

    def get_95th99th_percentile(self):
        # type: () -> (int, int)
        parr = self.get_percentile([0.95, 0.99])
        return parr[0], parr[1]

    def clear(self):
        self._values = []


class ReqStats:
    """
    Statistics per type of request.
    """

    def __init__(self, profile):
        # type: (StatsProfile) -> None
        self._httpRequestCount = 0
        self._errors = 0
        self._reqSizeMin = sys.maxsize
        self._reqSizeMax = 0
        self._reqSizeSum = 0
        self._resSizeMin = sys.maxsize
        self._resSizeMax = 0
        self._resSizeSum = 0
        self._retryAuthCount = 0
        self._retryThrottleCount = 0
        self._retryCount = 0
        self._retryDelayMs = 0
        self._rateLimitDelayMs = 0
        self._requestLatencyMin = sys.maxsize
        self._requestLatencyMax = 0
        self._requestLatencySum = 0
        if profile.value >= StatsProfile.MORE.value:
            self._requestLatencyPercentile = Percentile()
        else:
            self._requestLatencyPercentile = None

    def observe(self, error, retries, retry_delay,
                rate_limit_delay, auth_count, throttle_count,
                req_size, res_size, network_latency):
        # type: (bool, int, int, int, int, int, int, int, int) -> None
        self._httpRequestCount += 1
        self._retryCount += retries
        self._retryDelayMs += retry_delay
        self._retryAuthCount += auth_count
        self._retryThrottleCount += throttle_count
        self._rateLimitDelayMs += rate_limit_delay

        if error:
            self._errors += 1
        else:
            self._reqSizeMin = min(self._reqSizeMin, req_size)
            self._reqSizeMax = max(self._reqSizeMax, req_size)
            self._reqSizeSum += req_size
            self._resSizeMin = min(self._resSizeMin, res_size)
            self._resSizeMax = max(self._resSizeMax, res_size)
            self._resSizeSum += res_size
            self._requestLatencyMin = \
                min(self._requestLatencyMin, network_latency)
            self._requestLatencyMax = \
                max(self._requestLatencyMax, network_latency)
            self._requestLatencySum += network_latency
            if self._requestLatencyPercentile is not None:
                self._requestLatencyPercentile.add_value(network_latency)

    def to_json(self, request_name, req_array):
        # type: (str, []) -> None
        if self._httpRequestCount > 0:
            req = {"name": request_name}
            self.to_map_value(req)
            req_array.append(req)

    def to_map_value(self, map_value):
        map_value["httpRequestCount"] = self._httpRequestCount
        map_value["errors"] = self._errors

        retry = {
            "count": self._retryCount,
            "delayMs": self._retryDelayMs,
            "authCount": self._retryAuthCount,
            "throttleCount": self._retryThrottleCount
        }
        map_value["retry"] = retry
        map_value["rateLimitDelayMs"] = self._rateLimitDelayMs

        if self._requestLatencyMax > 0:
            latency = {
                "min": self._requestLatencyMin,
                "max": self._requestLatencyMax,
                "avg": self._requestLatencySum / (self._httpRequestCount - self._errors)
            }
            if self._requestLatencyPercentile is not None:
                (p95th, p99th) = \
                    self._requestLatencyPercentile.get_95th99th_percentile()
                latency["95th"] = p95th
                latency["99th"] = p99th
            map_value["httpRequestLatencyMs"] = latency

        if self._reqSizeMax > 0:
            reqSize = {
                "min": self._reqSizeMin,
                "max": self._reqSizeMax,
                "avg": self._reqSizeSum / (self._httpRequestCount - self._errors)
            }
            map_value["requestSize"] = reqSize

        if self._resSizeMax > 0:
            resSize = {
                "min": self._resSizeMin,
                "max": self._resSizeMax,
                "avg": self._resSizeSum / (self._httpRequestCount - self._errors)
            }
            map_value["resultSize"] = resSize

    def clear(self):
        self._httpRequestCount = 0
        self._errors = 0
        self._reqSizeMin = sys.maxsize
        self._reqSizeMax = 0
        self._reqSizeSum = 0
        self._resSizeMin = sys.maxsize
        self._resSizeMax = 0
        self._resSizeSum = 0
        self._retryAuthCount = 0
        self._retryThrottleCount = 0
        self._retryCount = 0
        self._retryDelayMs = 0
        self._rateLimitDelayMs = 0
        self._requestLatencyMin = sys.maxsize
        self._requestLatencyMax = 0
        self._requestLatencySum = 0
        if self._requestLatencyPercentile is not None:
            self._requestLatencyPercentile.clear()


class Stats:
    """
    Implements all the statistics.
    """
    _stats_control = None  # type: StatsControl
    _timer = None         # type: RepeatedTimer
    _requests = None      # type: Dict[str, ReqStats]

    _request_names = ["Delete", "Get", "GetIndexes", "GetTable",
                      "ListTables", "MultiDelete", "Prepare", "Put", "Query",
                      "System", "SystemStatus", "Table", "TableUsage",
                      "WriteMultiple", "Write"]

    def __init__(self, stats_control):
        # type: (StatsControl) -> None
        self._stats_control = stats_control

        interval = 10

        local_time = localtime()
        delay = (1000 * interval -
                 ((1000 * 60 * local_time.tm_min +
                   1000 * local_time.tm_sec) % (1000 * interval))) / 1000

        self._timer = RepeatedTimer(delay, interval, self.log_client_stats)

        self._start_time = datetime.utcnow()
        self._end_time = self._start_time
        self._extra_query_stats = None
        profile = self._stats_control.get_profile()
        if profile is not None and profile.value >= StatsProfile.ALL.value:
            self._extra_query_stats = ExtraQueryStats(profile)

        self._requests = {}
        for i in self._request_names:
            self._requests[i] = ReqStats(stats_control.get_profile())
        self.lock = Lock()

    @synchronized
    def log_client_stats(self):
        self.__log_client_stats()

    def __log_client_stats(self):
        handler = self._stats_control.get_stats_handler()
        log = self._stats_control.get_logger() is not None and \
                 self._stats_control.get_logger().isEnabledFor(INFO)

        if handler is not None or log:
            stats = self.__generate_stats()
            self.clear()

            if handler is not None:
                handler(stats)

            if log:
                if self._stats_control.get_pretty_print():
                    stats_str = pprint.pformat(stats)
                else:
                    stats_str = str(stats)
                self._stats_control.get_logger().info(StatsControl.LOG_PREFIX +
                                                      stats_str)

    def __generate_stats(self):
        self._end_time = datetime.utcnow()
        root = {
            "startTime": self._start_time.strftime("%Y-%m-%dT%H:%M:%SZ"),
            "endTime": self._end_time.strftime("%Y-%m-%dT%H:%M:%SZ"),
            "clientId": self._stats_control.get_id()
        }

        if self._extra_query_stats is not None:
            self._extra_query_stats.to_json(root)

        req_array = []
        for k in self._requests:
            self._requests[k].to_json(k, req_array)

        root["requests"] = req_array
        return root

    def clear(self):
        self._start_time = datetime.utcnow()
        self._end_time = self._start_time
        if self._extra_query_stats is not None:
            self._extra_query_stats.clear()

        for k in self._requests:
            self._requests[k].clear()

    @synchronized
    def shutdown(self):
        self.__log_client_stats()
        self._timer.stop()

    @synchronized
    def observe_error(self, request):
        self.__observe(request, True, -1, -1, -1)

    @synchronized
    def observe(self, request, error, req_size, res_size, network_latency):
        self.__observe(request, error, req_size, res_size,
                       network_latency)

    def __observe(self, request, error, req_size, res_size,
                  network_latency):
        reqStr = request.get_request_name()
        req_stat = self._requests.get(reqStr)
        if req_stat is None:
            req_stat = ReqStats(self._stats_control.get_profile())
            self._requests[reqStr] = req_stat

        auth_count = 0
        throttle_count = 0
        if request.get_retry_stats() is not None:
            auth_count = request.get_retry_stats().get_num_exceptions(
                AuthenticationException)
            auth_count += request.get_retry_stats().get_num_exceptions(
                SecurityInfoNotReadyException)
            throttle_count = request.get_retry_stats().get_num_exceptions(
                ThrottlingException)

        req_stat.observe(error, request.get_num_retries(),
                         request.get_retry_delay_ms(),
                         request.get_rate_limit_delayed_ms(), auth_count,
                         throttle_count, req_size, res_size, network_latency)

        from . import QueryRequest

        if self._extra_query_stats is not None and \
                isinstance(request, QueryRequest) and \
                self._stats_control.get_profile().value >= \
                StatsProfile.ALL.value:
            self._extra_query_stats.observe_q_rec(request, error,
                                                  request.get_num_retries(),
                                                  request.get_retry_delay_ms(),
                                                  request.
                                                  get_rate_limit_delayed_ms(),
                                                  auth_count, throttle_count,
                                                  req_size, res_size,
                                                  network_latency)

    @synchronized
    def observe_query(self, query_request):
        if self._extra_query_stats is None and \
                self._stats_control.get_profile().value >= \
                StatsProfile.ALL.value:
            self._extra_query_stats = \
                ExtraQueryStats(self._stats_control.get_profile())
        if self._extra_query_stats is not None and \
                self._stats_control.get_profile().value >= \
                StatsProfile.ALL.value:
            self._extra_query_stats.observe_query(query_request)



#
# Copyright (c) 2018, 2022 Oracle and/or its affiliates. All rights reserved.
#
# Licensed under the Universal Permissive License v 1.0 as shown at
#  https://oss.oracle.com/licenses/upl/
#

from .exception import AuthenticationException
from .kv import StoreAccessTokenProvider

__all__ = ['AuthenticationException',
           'StoreAccessTokenProvider'
           ]



#
# Copyright (c) 2018, 2022 Oracle and/or its affiliates. All rights reserved.
#
# Licensed under the Universal Permissive License v 1.0 as shown at
#  https://oss.oracle.com/licenses/upl/
#

from borneo.exception import NoSQLException


class AuthenticationException(NoSQLException):
    """
    On-premise only.

    This exception is thrown when use StoreAccessTokenProvider in following
    cases:

        Authentication information was not provided in the request header.\n
        The authentication session has expired. By default
        :py:class:`StoreAccessTokenProvider` will automatically retry
        authentication operation based on its authentication information.
    """

    def __init__(self, message, cause=None):
        super(AuthenticationException, self).__init__(message, cause)
